import os
from functools import partial
from typing import Union, TYPE_CHECKING

import numpy as np
import pandas as pd

from mewpy.io.dto import VariableRecord, DataTransferObject, FunctionTerm
from mewpy.germ.algebra import Expression, NoneAtom
from mewpy.germ.models import RegulatoryModel
from .engine import Engine
from .engines_utils import build_symbolic, expression_warning, csv_warning

if TYPE_CHECKING:
    from mewpy.germ.models import RegulatoryModel, Model, MetabolicModel


class BooleanRegulatoryCSV(Engine):
    def __init__(self, io, config, model=None):
        """
        Engine for CSV and TXT Boolean-based Regulatory files
        """
        super().__init__(io, config, model)

    @property
    def model_type(self):
        return 'regulatory'

    @property
    def model(self):

        if self._model is None:
            identifier = self.get_identifier()

            return RegulatoryModel(identifier=identifier)

        return self._model

    def build_data_frame(self):

        sep = self.config.get('sep', ',')
        id_col = self.config.get('id_col', 0)
        rule_col = self.config.get('rule_col', 1)
        aliases_cols = self.config.get('aliases_cols', [])
        header = self.config.get('header', None)
        filter_nan = self.config.get('filter_nan', False)

        names = {id_col: 'ids', rule_col: 'rules'}
        names.update({j: 'aliases_' + str(i + 1) for i, j in enumerate(aliases_cols)})

        try:
            df = pd.read_csv(self.io, sep=sep, header=header)

        except BaseException as exc:
            self.clean()
            self.close()
            raise exc

        cols = []

        for j, col in enumerate(df.columns):
            if j in names:
                cols.append(names[j])
            else:
                del df[col]

        df.columns = cols
        df.index = df.loc[:, 'ids']

        if filter_nan:

            df = df.dropna(subset=['rules'])

        else:

            df = df.replace(np.nan, '', regex=True)

        self.dto.data_frame = df

        self.dto.aliases_columns = [col for col in df.columns if col != 'ids' and col != 'rules']

    def get_identifier(self):

        if os.path.exists(self.io):
            _, identifier = os.path.split(self.io)
            return os.path.splitext(identifier)[0]

        return 'model'

    def open(self, mode='r'):

        self._dto = DataTransferObject()

        if not os.path.exists(self.io):
            raise OSError(f'{self.io} is not a valid input. Provide the path or file handler')

        self.dto.id = self.get_identifier()

    def parse(self):

        if self.dto is None:
            raise OSError('File is not open')

        if self.dto.id is None:
            raise OSError('File is not open')

        # -----------------------------------------------------------------------------
        # CSV/TXT to pandas dataframe
        # -----------------------------------------------------------------------------

        self.build_data_frame()

        for target in self.dto.data_frame.index:

            # -----------------------------------------------------------------------------
            # Target
            # -----------------------------------------------------------------------------

            target_id = target.replace(' ', '')

            target_aliases = self.dto.data_frame.loc[target, self.dto.aliases_columns]

            target_record = VariableRecord(id=target_id,
                                           name=target_id,
                                           aliases=set(target_aliases))

            self.variables[target_id].add('target')

            self.dto.targets[target_id] = target_record

            # -----------------------------------------------------------------------------
            # Regulators
            # -----------------------------------------------------------------------------

            symbolic, warning = build_symbolic(self.dto.data_frame.loc[target, 'rules'])

            if warning:
                self.warnings.append(partial(expression_warning, warning))

            regulators = {}

            for symbol in symbolic.atoms(symbols_only=True):
                self.variables[symbol.name].add('regulator')

                regulator_record = VariableRecord(id=symbol.name,
                                                  name=symbol.name,
                                                  aliases={symbol.value, symbol.name})

                regulators[symbol.name] = regulator_record

                self.dto.regulators[symbol.name] = regulator_record

            # -----------------------------------------------------------------------------
            # Function terms
            # -----------------------------------------------------------------------------
            function_terms = {'default_term': FunctionTerm(id='default_term', symbolic=NoneAtom(), coefficient=0),
                              'active_term': FunctionTerm(id='active_term', symbolic=symbolic, coefficient=1)}

            # -----------------------------------------------------------------------------
            # Interaction
            # -----------------------------------------------------------------------------

            interaction_id = f'{target_id}_interaction'

            interaction_record = VariableRecord(id=interaction_id,
                                                name=interaction_id,
                                                aliases={target_id},
                                                target=target_record,
                                                function_terms=function_terms,
                                                regulators=regulators)

            self.dto.interactions[interaction_id] = interaction_record

            self.variables[interaction_id].add('interaction')

    def read(self,
             model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
             variables=None):

        if not model:
            model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = self.model

        if not variables:
            variables = self.variables

        if self.dto.id:
            model._id = self.dto.id

        processed_regulators = set()

        for interaction_id, interaction_record in self.dto.interactions.items():

            target_record = interaction_record.target

            target, warning = target_record.to_variable(model=model,
                                                        types=variables.get(target_record.id, {'target'}),
                                                        name=target_record.name,
                                                        aliases=target_record.aliases)

            if warning:
                self.warnings.append(partial(csv_warning, warning))

            regulators_records = interaction_record.regulators

            regulators = {}

            for regulator_id, regulator_record in regulators_records.items():

                regulator, warning = regulator_record.to_variable(model=model,
                                                                  types=variables.get(regulator_id, {'regulator'}),
                                                                  name=regulator_record.name,
                                                                  aliases=regulator_record.aliases)

                if warning:
                    self.warnings.append(partial(csv_warning, warning))

                regulators[regulator_id] = regulator

                processed_regulators.add(regulator_id)

            regulatory_events = {}

            for func_term in interaction_record.function_terms.values():

                expression_regulators = {symbol.name: regulators[symbol.name]
                                         for symbol in func_term.symbolic.atoms(symbols_only=True)}

                regulatory_events[func_term.coefficient] = Expression(symbolic=func_term.symbolic,
                                                                      variables=expression_regulators)

            interaction, warning = interaction_record.to_variable(model=model,
                                                                  types=variables.get(interaction_id, {'interaction'}),
                                                                  name=interaction_record.name,
                                                                  aliases=interaction_record.aliases,
                                                                  target=target,
                                                                  regulatory_events=regulatory_events)

            if warning:
                self.warnings.append(partial(csv_warning, warning))

            model.add(interaction, comprehensive=True)

        if len(processed_regulators) != len(self.dto.regulators):

            for regulator_id, regulator_record in self.dto.regulators.items():

                if regulator_id not in processed_regulators:

                    regulator, warning = regulator_record.to_variable(model=model,
                                                                      types=variables.get(regulator_id, {'regulator'}),
                                                                      name=regulator_record.name,
                                                                      aliases=regulator_record.aliases)

                    if warning:
                        self.warnings.append(partial(csv_warning, warning))

                    model.add(regulator)

        return model

    def write(self):
        pass

    def close(self):

        if hasattr(self.io, 'close'):
            self.io.close()

    def clean(self):
        self._dto = None



from pathlib import Path
from typing import Type, Union, TYPE_CHECKING

from .builder import Builder
from .engines import Engines

if TYPE_CHECKING:
    from .engines.engine import Engine
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel
    from cobra import Model as CobraModel
    from reframed import CBModel as ReframedModel


class Writer(Builder):
    """
    The Writer is just a wrapper for an engine. It just wraps configurations and provides the simple API write
    Yet, this object is the user interface to write files, file handlers, cbm models from cobrapy and reframed

    Wrapping is accomplished using the composition pattern

    The only difference between the builders and the engines is that engines write the model in four steps:
                - open
                - write
                - close
                - clean

    Builders, on the other hand, just write and thus performing all operations at once.
    This is important for writing multiple files out of a single model.

    For writing files in stages, multiple writers must be created and the director must be used to merge all the
    writers into a single model
    """
    def __init__(self,
                 engine: Union[Type['Engine'], Engines],
                 io: Union[str, Path, 'CobraModel', 'ReframedModel'],
                 model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                 config: dict = None):

        """
        Write a given model type into a file of any type.

        It currently supports xml, sbml and json. It also supports writing of cobrapy and reframed models.

        An engine listed in the Engines enumerator of the mewpy.io.engines module must be provided. This engine
        determines the file type.

        :param engine: A valid Engine listed in the Engines enumerator of the mewpy.io.engines.
        It will handle file writing
        :param io: A valid string path or IO is acceptable.
        Alternatively, it can also be provided a cobrapy and reframed model that will be filled
        :param model: A valid metabolic, regulatory or both GERM model
        :param config: Dictionary with additional configurations

        """
        if not engine:
            raise ValueError('Nothing to write. Please provide an engine')

        if not io:
            raise ValueError('Nothing to write. Please provide a path, file handler or model')

        if not model:
            raise ValueError('Nothing to write. Please provide a GERM model')

        if isinstance(io, Path):
            io = str(io)

        if not isinstance(engine, Engines):

            old_engine = engine

            engine = Engines.get(old_engine)

            if engine is None:
                raise ValueError(f'{old_engine} is not supported. See available engines at {Engines}')

        engine = engine.value

        if not config:
            config = {}

        super(Writer, self).__init__()

        self._engine = engine(io=io, config=config, model=model)

    def __enter__(self):

        self._context = True

        self.engine.open()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        self._context = False

        self.engine.close()
        self.engine.clean()

    # to understand this read method consult the director, builder and reader init stubs
    def write(self, warnings: bool = True):
        """
        Writing a metabolic, regulatory or GERM model to a file type
        (e.g. sbml, json, cobrapy, reframed, etc).
        The file is closed upon writing or failure

        :param warnings: Whether to launch warnings found during reading
        """

        if self._context:

            self.engine.write()

        else:
            self.engine.open()

            self.engine.write()

            self.engine.close()
            self.engine.clean()

        if warnings:

            for warning in self.warnings:
                warning()



""" Adapted from Daniel Machado's REFRAMED"""
from .solver import Solver, VarType, Parameter, default_parameters
from .solution import Solution, Status
from gurobipy import Model as GurobiModel, GRB, quicksum
from math import inf
from warnings import warn


def infinity_fix(val):
    if val == inf:
        return GRB.INFINITY
    elif val == -inf:
        return -GRB.INFINITY
    else:
        return val


status_mapping = {
    GRB.OPTIMAL: Status.OPTIMAL,
    GRB.UNBOUNDED: Status.UNBOUNDED,
    GRB.INFEASIBLE: Status.INFEASIBLE,
    GRB.INF_OR_UNBD: Status.INF_OR_UNB
}

vartype_mapping = {
    VarType.BINARY: GRB.BINARY,
    VarType.INTEGER: GRB.INTEGER,
    VarType.CONTINUOUS: GRB.CONTINUOUS
}

parameter_mapping = {
    Parameter.TIME_LIMIT: GRB.Param.TimeLimit,
    Parameter.FEASIBILITY_TOL: GRB.Param.FeasibilityTol,
    Parameter.INT_FEASIBILITY_TOL: GRB.Param.IntFeasTol,
    Parameter.OPTIMALITY_TOL: GRB.Param.OptimalityTol,
    Parameter.MIP_ABS_GAP: GRB.Param.MIPGapAbs,
    Parameter.MIP_REL_GAP: GRB.Param.MIPGap,
    Parameter.POOL_SIZE: GRB.Param.PoolSolutions,
    Parameter.POOL_GAP: GRB.Param.PoolGap
}


class GurobiSolver(Solver):
    """ Implements the gurobi solver interface. """

    def __init__(self, model=None):
        Solver.__init__(self)
        self.problem = GurobiModel()
        self.set_logging(False)
        self.set_parameters(default_parameters)
        if model:
            self.build_problem(model)

    def add_variable(self, var_id, lb=-inf, ub=inf, vartype=VarType.CONTINUOUS, update=True):
        """ Add a variable to the current problem.

        Arguments:
            var_id (str): variable identifier
            lb (float): lower bound
            ub (float): upper bound
            vartype (VarType): variable type (default: CONTINUOUS)
            update (bool): update problem immediately (default: True)
        """

        lb = infinity_fix(lb)
        ub = infinity_fix(ub)

        if var_id in self.var_ids:
            var = self.problem.getVarByName(var_id)
            var.setAttr('lb', lb)
            var.setAttr('ub', ub)
            var.setAttr('vtype', vartype_mapping[vartype])
        else:
            self.problem.addVar(name=var_id, lb=lb, ub=ub, vtype=vartype_mapping[vartype])
            self.var_ids.append(var_id)

        if update:
            self.problem.update()

    def set_variable_bounds(self, var_id, lb, ub):
        """Modify a variable bounds

        Args:
            var_id (str): [description]
            lb (float): lower bound
            ub (float): upper bound
        """
        var = self.problem.getVarByName(var_id)
        if lb:
            var.lb = lb
        if ub:
            var.ub = ub

    def add_constraint(self, constr_id, lhs, sense='=', rhs=0, update=True):
        """ Add a constraint to the current problem.

        Arguments:
            constr_id (str): constraint identifier
            lhs (dict): variables and respective coefficients
            sense (str): constraint sense (any of: '<', '=', '>'; default '=')
            rhs (float): right-hand side of equation (default: 0)
            update (bool): update problem immediately (default: True)
        """

        grb_sense = {'=': GRB.EQUAL,
                     '<': GRB.LESS_EQUAL,
                     '>': GRB.GREATER_EQUAL}

        if constr_id in self.constr_ids:
            constr = self.problem.getConstrByName(constr_id)
            self.problem.remove(constr)

        expr = quicksum(coeff * self.problem.getVarByName(r_id) for r_id, coeff in lhs.items() if coeff)

        self.problem.addLConstr(expr, grb_sense[sense], rhs, constr_id)
        self.constr_ids.append(constr_id)

        if update:
            self.problem.update()

    def remove_variable(self, var_id):
        """ Remove a variable from the current problem.

        Arguments:
            var_id (str): variable identifier
        """
        self.remove_variables([var_id])

    def remove_variables(self, var_ids):
        """ Remove variables from the current problem.

        Arguments:
            var_ids (list): variable identifiers
        """

        for var_id in var_ids:
            if var_id in self.var_ids:
                self.problem.remove(self.problem.getVarByName(var_id))
                self.var_ids.remove(var_id)

    def remove_constraint(self, constr_id):
        """ Remove a constraint from the current problem.

        Arguments:
            constr_id (str): constraint identifier
        """
        self.remove_constraints([constr_id])

    def remove_constraints(self, constr_ids):
        """ Remove constraints from the current problem.

        Arguments:
            constr_ids (list): constraint identifiers
        """

        for constr_id in constr_ids:
            if constr_id in self.constr_ids:
                self.problem.remove(self.problem.getConstrByName(constr_id))
                self.constr_ids.remove(constr_id)

    def update(self):
        """ Update internal structure. Used for efficient lazy updating. """
        self.problem.update()

    def set_objective(self, linear=None, quadratic=None, minimize=True):
        """ Set a predefined objective for this problem.

        Args:
            linear (dict): linear coefficients (optional)
            quadratic (dict): quadratic coefficients (optional)
            minimize (bool): solve a minimization problem (default: True)

        Notes:
            Setting the objective is optional. It can also be passed directly when calling **solve**.

        """

        lin_obj = []
        quad_obj = []

        if linear:

            if isinstance(linear, str):
                lin_obj = [1.0 * self.problem.getVarByName(linear)]
                if linear not in self.var_ids:
                    warn(f"Objective variable not previously declared: {linear}")
            else:
                lin_obj = []
                for r_id, val in linear.items():
                    if r_id not in self.var_ids:
                        warn(f"Objective variable not previously declared: {r_id}")
                    elif val != 0:
                        lin_obj.append(val * self.problem.getVarByName(r_id))

        if quadratic:
            quad_obj = []
            for (r_id1, r_id2), val in quadratic.items():
                if r_id1 not in self.var_ids:
                    warn(f"Objective variable not previously declared: {r_id1}")
                elif r_id2 not in self.var_ids:
                    warn(f"Objective variable not previously declared: {r_id2}")
                elif val != 0:
                    quad_obj.append(val * self.problem.getVarByName(r_id1) * self.problem.getVarByName(r_id2))

        obj_expr = quicksum(quad_obj + lin_obj)
        sense = GRB.MINIMIZE if minimize else GRB.MAXIMIZE

        self.problem.setObjective(obj_expr, sense)

    def solve(self, linear=None, quadratic=None, minimize=None, model=None, constraints=None, get_values=True,
              shadow_prices=False, reduced_costs=False, pool_size=0, pool_gap=None):
        """ Solve the optimization problem.

        Arguments:
            linear (str or dict): linear coefficients (or a single variable to optimize)
            quadratic (dict): quadratic objective (optional)
            minimize (bool): solve a minimization problem (default: True)
            model (CBModel): model (optional, leave blank to reuse previous model structure)
            constraints (dict): additional constraints (optional)
            get_values (bool or list): set to false for speedup if you only care about the objective (default: True)
            shadow_prices (bool): return shadow prices if available (default: False)
            reduced_costs (bool): return reduced costs if available (default: False)
            pool_size (int): calculate solution pool of given size (only for MILP problems)
            pool_gap (float): maximum relative gap for solutions in pool (optional)

        Returns:
            Solution: solution
        """

        if model:
            self.build_problem(model)

        problem = self.problem

        if constraints:
            old_constraints = {}
            for r_id, x in constraints.items():
                lb, ub = x if isinstance(x, tuple) else (x, x)
                if r_id in self.var_ids:
                    lpvar = problem.getVarByName(r_id)
                    old_constraints[r_id] = (lpvar.lb, lpvar.ub)
                    lpvar.lb = infinity_fix(lb)
                    lpvar.ub = infinity_fix(ub)
                else:
                    warn(f"Constrained variable '{r_id}' not previously declared")
            problem.update()

        self.set_objective(linear, quadratic, minimize)

        # run the optimization
        if pool_size <= 1:

            problem.optimize()

            status = status_mapping.get(problem.status, Status.UNKNOWN)
            message = str(problem.status)

            if status == Status.OPTIMAL:
                fobj = problem.ObjVal
                values, s_prices, r_costs = None, None, None

                if get_values:
                    try:
                        get_values = list(get_values)
                        values = {r_id: problem.getVarByName(r_id).X for r_id in get_values}
                    except Exception:
                        values = {r_id: problem.getVarByName(r_id).X for r_id in self.var_ids}

                if shadow_prices:
                    s_prices = {m_id: problem.getConstrByName(m_id).Pi for m_id in self.constr_ids}

                if reduced_costs:
                    r_costs = {r_id: problem.getVarByName(r_id).RC for r_id in self.var_ids}

                solution = Solution(status, message, fobj, values, s_prices, r_costs)
            else:
                solution = Solution(status, message)

        else:

            problem.setParam(GRB.Param.PoolSearchMode, 2)
            self.set_parameter(Parameter.POOL_SIZE, pool_size)

            if pool_gap:
                self.set_parameter(Parameter.POOL_GAP, pool_gap)

            problem.optimize()

            status = status_mapping.get(problem.status, Status.UNKNOWN)

            if status == Status.OPTIMAL or status == Status.UNKNOWN:
                solution = self.get_solution_pool()
            else:
                solution = []

        # restore values of temporary constraints
        if constraints:
            for r_id, (lb, ub) in old_constraints.items():
                lpvar = problem.getVarByName(r_id)
                lpvar.lb, lpvar.ub = lb, ub
            problem.update()

        return solution

    def get_solution_pool(self, get_values=True):
        """ Return a solution pool for MILP problems.
        Must be called after using solve with pool_size argument > 0.

        Arguments:
            get_values (bool or list): set to false for speedup if you only care about the objective (default: True)

        Returns:
            list: list of Solution objects

        """
        solutions = []

        for i in range(self.problem.SolCount):
            self.problem.setParam(GRB.param.SolutionNumber, i)
            obj = self.problem.PoolObjVal
            if get_values:
                try:
                    get_values = list(get_values)
                    values = {r_id: self.problem.getVarByName(r_id).Xn for r_id in get_values}
                except Exception:
                    values = {r_id: self.problem.getVarByName(r_id).Xn for r_id in self.var_ids}
            else:
                values = None
            sol = Solution(fobj=obj, values=values)
            solutions.append(sol)

        return solutions

    def set_parameter(self, parameter, value):
        """ Set a parameter value for this optimization problem

        Arguments:
            parameter (Parameter): parameter type
            value (float): parameter value
        """

        if parameter in parameter_mapping:
            grb_param = parameter_mapping[parameter]
            self.problem.setParam(grb_param, value)
        else:
            raise Exception('Parameter unknown (or not yet supported).')

    def set_logging(self, enabled=False):
        """ Enable or disable log output:

        Arguments:
            enabled (bool): turn logging on (default: False)
        """

        self.problem.setParam('OutputFlag', 1 if enabled else 0)

    def write_to_file(self, filename):
        """ Write problem to file:

        Arguments:
            filename (str): file path
        """

        self.problem.write(filename)

    def change_coefficients(self, coefficients):
        """Changes variables coefficients in constraints

        :param coefficients: A list of tuples (constraint name, variable name, new value)
        :type coefficients: list
        """
        for c_id, v_id, x in coefficients:
            constraint = self.problem.getConstrByName(c_id)
            variable = self.problem.getVarByName(v_id)
            self.problem.chgCoeff(constraint, variable, x)



import re
from io import StringIO
from token import NAME, OP, NUMBER
from tokenize import generate_tokens, untokenize
from typing import List

from .symbolic import NoneAtom, Symbolic
from .algebra_constants import (BOOLEAN_STATES,
                                BOOLEAN_OPERATORS,
                                TRUE,
                                FALSE,
                                RELATIONAL_STATES,
                                RELATIONAL_OPERATORS,
                                RELATIONAL_EQUAL_OPERATORS,
                                GLOBAL_RELATIONAL_EQUAL_OPERATORS,
                                GLOBAL_RELATIONAL_OPERATORS,
                                BOOLEAN_ESCAPE_CHARS,
                                RELATIONAL_ESCAPE_CHARS,
                                GLOBAL_MEWPY_OPERATORS)

_relational_ops = set()
_relational_ops.update(GLOBAL_RELATIONAL_OPERATORS)
_relational_ops.update(GLOBAL_RELATIONAL_EQUAL_OPERATORS)


class ExpressionParser:

    def __init__(self,
                 expression=None,

                 stringify_expression='',
                 parsed_expression='',
                 symbolic_expression='',
                 tokenized_expression=None,

                 tokens=None,
                 filters=None,
                 transformations=None,
                 escape_chars=None,
                 replaces=None,

                 symbols=None,
                 aliases=None,

                 is_boolean=False,
                 is_true=False,
                 is_false=False,
                 is_and=False,
                 is_or=False,
                 is_not=False,

                 is_relational=False,
                 is_equal=False,
                 is_not_equal=False,
                 is_inequality=False,
                 is_greater=False,
                 is_greater_equal=False,
                 is_less=False,
                 is_less_equal=False,

                 is_numeric=False,
                 is_integer=False,
                 is_float=False,
                 is_one=False,
                 is_zero=False,

                 is_symbol=False,
                 is_atom=False,

                 is_none=False):

        """
        Internal use only!

        The ExpressionParser object is to be used only by the next parsing, filtering, transformation
        and processing methods.
        It infers the algebra expression nature, and holds corresponding transforming, filtering and replacing methods
        and dictionaries to be applied to the raw expression.

        """

        if aliases is None:
            aliases = {}

        if symbols is None:
            symbols = {}

        if replaces is None:
            replaces = {}

        if escape_chars is None:
            escape_chars = {}

        if tokens is None:
            tokens = []

        if filters is None:
            filters = []

        if transformations is None:
            transformations = []

        if tokenized_expression is None:
            tokenized_expression = []

        self.expression = expression

        self.stringify_expression = stringify_expression
        self.parsed_expression = parsed_expression
        self.symbolic_expression = symbolic_expression
        self.tokenized_expression = tokenized_expression

        self.tokens = tokens
        self.filters = filters
        self.transformations = transformations
        self.escape_chars = escape_chars
        self.replaces = replaces

        self.symbols = symbols
        self.aliases = aliases

        self.is_boolean = is_boolean
        self.is_true = is_true
        self.is_false = is_false
        self.is_and = is_and
        self.is_or = is_or
        self.is_not = is_not

        self.is_relational = is_relational
        self.is_equal = is_equal
        self.is_not_equal = is_not_equal
        self.is_inequality = is_inequality
        self.is_greater = is_greater
        self.is_greater_equal = is_greater_equal
        self.is_less = is_less
        self.is_less_equal = is_less_equal

        self.is_numeric = is_numeric
        self.is_integer = is_integer
        self.is_float = is_float
        self.is_one = is_one
        self.is_zero = is_zero

        self.is_symbol = is_symbol
        self.is_atom = is_atom

        self.is_none = is_none

    def build(self):

        for token in self.tokenized_expression:

            if token.lower() in BOOLEAN_OPERATORS:

                if self.is_boolean:
                    continue

                self.is_boolean = True
                diff = set(boolean_filters) - set(self.filters)
                self.filters = self.filters + list(diff)
                diff = set(boolean_transformations) - set(self.transformations)
                self.transformations = self.transformations + list(diff)
                self.escape_chars.update(BOOLEAN_ESCAPE_CHARS)

            elif token.lower() in RELATIONAL_OPERATORS or token.lower() in RELATIONAL_EQUAL_OPERATORS:

                if self.is_relational:
                    continue

                self.is_relational = True
                diff = set(relational_filters) - set(self.filters)
                self.filters = self.filters + list(diff)
                diff = set(relational_transformations) - set(self.transformations)
                self.transformations = self.transformations + list(diff)
                self.escape_chars.update(RELATIONAL_ESCAPE_CHARS)

            elif token.lower() in BOOLEAN_STATES or token.lower() in RELATIONAL_STATES:

                if self.is_one or self.is_true or self.is_false or self.is_zero:
                    continue

                bool_state = BOOLEAN_STATES.get(token.lower(), token.lower())

                if bool_state == TRUE:
                    self.is_one = True
                    self.is_true = True

                if bool_state == FALSE:
                    self.is_zero = True
                    self.is_false = True

                diff = set(state_filters) - set(self.filters)
                self.filters = self.filters + list(diff)
                diff = set(state_transformations) - set(self.transformations)
                self.transformations = self.transformations + list(diff)

            else:

                if self.is_symbol:
                    continue

                self.is_symbol = True
                diff = set(symbol_filters) - set(self.filters)
                self.filters = self.filters + list(diff)
                diff = set(symbol_transformations) - set(self.transformations)
                self.transformations = self.transformations + list(diff)
                self.escape_chars.update({**BOOLEAN_ESCAPE_CHARS, **RELATIONAL_ESCAPE_CHARS})

        if not self.is_boolean or self.is_relational or self.is_one or self.is_true or \
                self.is_zero or self.is_false or self.is_symbol:
            self.filters = all_filters
            self.transformations = all_transformations
            self.escape_chars = {**BOOLEAN_ESCAPE_CHARS, **RELATIONAL_ESCAPE_CHARS}


def tokenize(rule: str) -> List[str]:
    """
    Tokenizes a stringify expression.
    :param rule: stringify expression as string
    :return: it returns all tokens of the expression
    """
    return list(filter(lambda x: x != '', rule.replace('(', ' ( ').replace(')', ' ) ').split(' ')))


def escape_chars_filter(expression: ExpressionParser):
    """
    For a given expression it parses out the special chars by replacing them with the corresponding values.
    The global dictionary available in the algebra constants module is used according to the expression type.
    This global dictionary can be altered for adding or removing special chars.
    :param expression: the ExpressionParser object
    """

    for escape_char, replace in expression.escape_chars.items():

        if escape_char in expression.parsed_expression:
            expression.replaces[replace] = escape_char
            expression.parsed_expression = expression.parsed_expression.replace(escape_char, replace)


def digit_filter(expression):
    """
    For a given regulatory self it checks if all regulatory variables start with a digit (str.is_digit()) and
    parses out the regulatory variable by adding the prefix '_dg_'

    """

    expression.parsed_expression = '(' + expression.parsed_expression + ')'
    regexp = re.compile(r'[^a-zA-Z|^_][0-9]+[a-zA-Z]|[^a-zA-Z|^_][0-9]+_')
    res = list(regexp.finditer(expression.parsed_expression))

    new_rule = ''

    last_nd = 0
    for i in range(len(res)):
        st = res[i].start() + 1
        nd = res[i].end()
        new_rule = new_rule + expression.parsed_expression[last_nd:st] + '_dg_' + expression.parsed_expression[
                                                                                  st:nd]
        last_nd = res[i].end()
        expression.replaces['_dg_'] = ''

    new_rule = new_rule + expression.parsed_expression[last_nd:]

    expression.parsed_expression = new_rule[1:-1]


def bitwise_filter(expression):
    """
    Transforms python boolean operators and, or and not, or other propositional logic syntax (e.g. greater
    than) to corresponding bitwise ones
    """

    bit_replace = {}

    if expression.is_boolean:
        bit_replace.update(BOOLEAN_OPERATORS)

    if expression.is_relational:
        bit_replace.update(RELATIONAL_EQUAL_OPERATORS)

    if expression.is_one or expression.is_true or expression.is_zero or expression.is_false:
        bit_replace.update(RELATIONAL_STATES)
        bit_replace.update(BOOLEAN_STATES)

    bit_ = {}

    for key, val in bit_replace.items():

        if key == val:
            continue

        if val in bit_:
            bit_[val].update({key, key.upper(), key.lower(), key.title()})
        else:
            bit_[val] = {key, key.upper(), key.lower(), key.title()}

    for bit_val, bools in bit_.items():

        for python_bool in bools:
            expression.parsed_expression = re.sub(r'\b{}\b'.format(python_bool), bit_val,
                                                  expression.parsed_expression)


def relational_filter(expression):
    """
    For a given boolean self (string) finds all relational expressions and encloses all of them with
    parenthesis and replaces the ambiguous operators

    """

    # regex to match a relational of any type: x>0; x<0; x>=0; x=>0;reverse order 0>x ...; with or without white
    # spaces ([a-zA-Z0-9_]+>[0-9]+)|([a-zA-Z0-9_]+\str>\str[0-9]+)

    regex_str = r'|'.join([r'([a-zA-Z0-9_]+' + regex + r'[0-9.]+)' + r'|' +
                           r'([a-zA-Z0-9_]+\s' + regex + r'\s[0-9.]+)' + r'|' +
                           r'([0-9.]+' + regex + r'[a-zA-Z0-9_]+)' + r'|' +
                           r'([0-9.]+\s' + regex + r'\s[a-zA-Z0-9_]+)'
                           for regex in _relational_ops])

    regexp = re.compile(regex_str)
    res = list(regexp.finditer(expression.parsed_expression))

    for match in set(map(lambda x: x.group(), res)):
        expression.parsed_expression = re.sub(r'\b{}\b'.format(match), '(' + match + ')',
                                              expression.parsed_expression)


def symbolic_transform(expression):
    """
    Heavily inspired by the talented people contributing for the impressive sympy package

    :param expression:
    :return:
    """

    result = []

    expression.tokens.append((None, None))  # so zip traverses all tokens

    for tok, next_token in zip(expression.tokens, expression.tokens[1:]):

        token_number, token_val = tok

        if token_number == NAME:

            name = token_val

            if name == 'False':
                result.extend([(NAME, 'BoolFalse'),
                               (OP, '('),
                               (NAME, repr(str(name))),
                               (OP, ')')])

            elif name == 'True':
                result.extend([(NAME, 'BoolTrue'),
                               (OP, '('),
                               (NAME, repr(str(name))),
                               (OP, ')')])

            elif name == 'None':
                result.extend([(NAME, 'NoneAtom'),
                               (OP, '('),
                               (NAME, repr(str(name))),
                               (OP, ')')])

            else:
                result.extend([
                    (NAME, 'Symbol'),
                    (OP, '('),
                    (NAME, repr(str(token_val))),
                    (OP, ')'),
                ])

        else:

            result.append((token_number, token_val))

    expression.tokens = result


# FIXME: parsing does not handle minus so far. BOOLEAN_ESCAPE_CHARS = {'-': '_dash_',
# FIXME: parsing does not handle scientific notation so far. BOOLEAN_ESCAPE_CHARS = {'-': '_dash_',
def numeric_transform(expression):
    """

    Heavily inspired by the talented people contributing for the impressive sympy package

    :param expression:
    :return:
    """

    result = []

    for token_number, token_val in expression.tokens:

        if token_number == NUMBER:

            number = token_val

            if number in ('1.0', '1'):

                seq = [(NAME, 'One'),
                       (OP, '('),
                       (NUMBER, repr(str(number))),
                       (OP, ')')]

            elif number in ('0.0', '0'):

                seq = [(NAME, 'Zero'),
                       (OP, '('),
                       (NUMBER, repr(str(number))),
                       (OP, ')')]

            elif '.' in number or (('e' in number or 'E' in number)
                                   and not (number.startswith('0x') or number.startswith('0X'))):

                seq = [(NAME, 'Float'),
                       (OP, '('),
                       (NUMBER, repr(str(number))),
                       (OP, ')')]

            else:

                seq = [(NAME, 'Integer'),
                       (OP, '('),
                       (NUMBER, number),
                       (OP, ')')]

            result.extend(seq)

        else:

            result.append((token_number, token_val))

    expression.tokens = result


all_filters = [escape_chars_filter, digit_filter, bitwise_filter, relational_filter]
all_transformations = [symbolic_transform, numeric_transform]

boolean_filters = [escape_chars_filter, digit_filter, bitwise_filter]
boolean_transformations = [symbolic_transform, numeric_transform]

relational_filters = [escape_chars_filter, digit_filter, bitwise_filter, relational_filter]
relational_transformations = [symbolic_transform, numeric_transform]

symbol_filters = [escape_chars_filter, digit_filter]
symbol_transformations = [symbolic_transform]

state_filters = [bitwise_filter]
state_transformations = [symbolic_transform, numeric_transform]


def apply_filters(expression, filters=None):
    if not filters:
        filters = expression.filters

    sorted_filters = []

    for _filter in sorted_filters:
        if callable(_filter):
            if _filter is escape_chars_filter:
                sorted_filters.insert(0, _filter)
            else:
                sorted_filters.append(_filter)

    for _filter in filters:
        _filter(expression)

    return expression


def apply_transformations(expression, transformations=None):
    """
    Heavily inspired by the talented people contributing for the impressive sympy package

    :param expression:
    :param transformations:
    :return:
    """

    if not transformations:
        transformations = expression.transformations

    sorted_transformations = []

    for transform in transformations:
        if callable(transform):
            if transform is symbolic_transform:
                sorted_transformations.insert(0, transform)
            else:
                sorted_transformations.append(transform)

    for transform in sorted_transformations:
        transform(expression)

    expression.symbolic_expression = untokenize(expression.tokens)

    return expression


def evaluate_expression(symbolic_expression, local_dict=None, global_dict=None):
    if not local_dict:
        local_dict = {}

    if not global_dict:
        global_dict = GLOBAL_MEWPY_OPERATORS

    return eval(symbolic_expression, global_dict, local_dict)


def parse_expression(expression: str) -> Symbolic:
    """
    Parsing an expression encoded into a string object to a Symbolic object that provides symbolic algebra evaluation
    with Symbolic-based symbols using regular python binary operators or custom functions

    For more details, consult the Expression object.

    :param expression: an expression as string object containing symbols and python binary (or not) operators
    :return: it returns a Symbolic object that allows for symbolic algebra evaluation
    """
    if not expression:

        expr_parser = ExpressionParser()
        expr_parser.expression = NoneAtom()
        expr_parser.is_none = True

    elif isinstance(expression, str):

        expr_parser = ExpressionParser(expression=None,
                                       stringify_expression=expression,
                                       parsed_expression=expression,
                                       tokenized_expression=tokenize(expression))

        expr_parser.build()

        expr_parser = apply_filters(expr_parser)

        expr_parser.tokenized_expression = tokenize(expr_parser.parsed_expression)

        input_code = StringIO(expr_parser.parsed_expression.strip())

        for token_number, token_val, _, _, _ in generate_tokens(input_code.readline):
            expr_parser.tokens.append((token_number, token_val))

        expr_parser = apply_transformations(expr_parser)

        if len(expr_parser.tokenized_expression) < 1:
            expr_parser.expression = NoneAtom()
            expr_parser.is_none = True

        else:

            expr_parser.expression = evaluate_expression(expr_parser.symbolic_expression)

    else:

        raise ValueError('Expression could not be parsed')

    for element in expr_parser.expression:

        if element.is_symbol:

            old_reg_name = ''.join(element.value)

            for replace, special_char in expr_parser.replaces.items():
                old_reg_name = old_reg_name.replace(replace, special_char)

            element.name = old_reg_name

    return expr_parser.expression



import warnings
from mewpy.model.kinetic import ODEModel
from mewpy.solvers import KineticConfigurations
from mewpy.simulation import get_simulator
from mewpy.simulation.simulation import Simulator
from mewpy.simulation.kinetic import KineticSimulation
from collections import OrderedDict
from warnings import warn
import pandas as pd
import numpy as np
from numpy.random import normal
from re import search
from tqdm import tqdm


warnings.filterwarnings('ignore', 'Timeout')


def sample(vmaxs, sigma=0.1):
    k = vmaxs.keys()
    f = np.exp(normal(0, sigma, len(vmaxs)))
    v = np.array(list(vmaxs.values()))
    r = list(v*f)
    return dict(zip(k, r))


class HybridSimulation:

    def __init__(self, kmodel, cbmodel, gDW=564.0, envcond=dict(),
                 mapping=dict(), t_points=[0, 1e9],
                 timeout=KineticConfigurations.SOLVER_TIMEOUT):

        if not isinstance(kmodel, ODEModel):
            raise ValueError('model is not an instance of ODEModel.')
        
        if not isinstance(cbmodel, Simulator):
            self.sim = get_simulator(cbmodel, envcond=envcond)
        else:
            self.sim = cbmodel

        self.kmodel = kmodel
        self.mapping = mapping
        self.t_points = t_points
        self.timeout = timeout
        self.models_verification()
        self.gDW = gDW

    def __getstate__(self):
        state = OrderedDict(self.__dict__.copy())
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def get_kinetic_model(self):
        return self.kmodel

    def get_mapping(self):
        return self.mapping

    def models_verification(self):
        """
        Function that verifies if it's possible to perform the Hibrid Simulation.
        """
        kmodel = self.get_kinetic_model()
        cbmodel = self.sim.model
        mapping = self.get_mapping()

        for k,v in mapping.items():
            if k not in kmodel.ratelaws.keys():
                raise ValueError(f"could not find reaction {k} in the kinetic model ")
            if v[0] not in cbmodel.reactions:
                raise ValueError(f"could not find reaction {v[0]} in the steady-state model ")
        
        return True
        
    def mapping_conversion(self, fluxes):
        """
        Function that converts the kinetic fluxes into constraint-based fluxes.

        :param fluxes: kinetic simulation fluxes
        :type fluxes: dict
        :return: kinetic fluxes compatible with the constraint-based model
        """
        mapping = self.get_mapping()
        flxs = dict()
        for k, value in fluxes.items():
            if k in mapping.keys():
                v = mapping[k]
                flxs[v[0]] = v[1]*value * 3600/self.gDW
        if len(flxs) != 0:
            return flxs
        else:
            raise warn('Mapping not done properly, please redo mapping')

    def mapping_bounds(self, lbs, ubs):
        """
        Function that converts the kinetic bounds into constraint-based flux bounds.

        :param lbs: kinetic lower bounds
        :type fluxes: dict
        :param ubs: kinetic upper bounds
        :type fluxes: dict
        :return: constraints
        """
        mapping = self.get_mapping()
        flxs = dict()
        for k, value in lbs.items():
            if k in mapping.keys():
                v = mapping[k]
                a = v[1]*value * 3600/self.gDW
                b = v[1]*ubs[k] * 3600/self.gDW
                flxs[v[0]] = (a, b) if a < b else (b, a)
        if len(flxs) != 0:
            return flxs
        else:
            raise warn('Mapping not done properly, please redo mapping')

    def nsamples(self, vmaxs, n=1, sigma=0.1):
        """
        Generates n fluxes samples varying vmax values on a log-norm distribtution
        with mean 0 and std sigma.

        """
        kmodel = self.get_kinetic_model()
        ksample = []
        ksim = KineticSimulation(model=kmodel, t_points=self.t_points, timeout=self.timeout)
        for _ in tqdm(range(n)):
            v = sample(vmaxs, sigma=sigma)
            try:
                res = ksim.simulate(parameters=v)
                if res.fluxes:
                    ksample.append(res.fluxes)
            except Exception as e:
                warn.warning(str(e))
        df = pd.DataFrame(ksample)
        # drop any NaN if exist
        df.dropna()
        return df
    
    
    def simulate(self, objective=None, 
                       initcond=None, 
                       parameters=None, 
                       constraints=None, 
                       method='pFBA'):
        """
        This method performs a phenotype simulation hibridizing a kinetic and a constraint-based model.

        :param objective: steady-state optimization objective.
        :type objective: dict, optional
        :param initcond:List of numbers in which the kinetic simulation fluxes will be scaled.
        :param initcond: list, optional
        :param parameters: Kinetic simulation parameters.
        :type parameters: dict, optional
        :param constraints: Constraint-based model simulation constraints.
        :type constraints: dict, optional
        :param method: the phenotype simulation method
        :type method: str. Default 'pFBA'
        :returns: Returns the solution of the hibridization.
        """
        mapp = self.models_verification()
        kmodel = self.get_kinetic_model()

        ksim = KineticSimulation(model=kmodel, t_points=self.t_points, timeout=self.timeout)
        result = ksim.simulate(parameters=parameters, initcon=initcond)
        fluxes = result.fluxes
        
        if constraints is None:
            constraints = dict()

        if not mapp:
            fluxes = self.mapping_conversion(fluxes)
        constraints.update(fluxes)
        if objective:
            solution = self.sim.simulate(objective= objective, method=method, constraints=constraints)
        else:
            solution = self.sim.simulate(method=method, constraints=constraints)
        return solution


    def simulate_distribution(self, df, q1=0.1, q2=0.9, objective=None, method='pFBA', constraints=None):
        """Runs a pFBA on the steady-state model with fluxes constrained to ranges between the q1-th and q2-th percentile
        of fluxes distributions sampled from the kinetic model.
        The kinetic flux distributions are provided as panda dataframes.     
        """
        const = dict()
        lbs = df.quantile(q1, axis=0).to_dict()
        ubs = df.quantile(q2, axis=0).to_dict()
        if constraints:
            const.update(constraints)
        k_const = self.mapping_bounds(lbs, ubs)
        const.update(k_const)
        if objective:
            solution = self.sim.simulate(method=method, constraints=const, objective=objective)
        else:
            solution = self.sim.simulate(method=method, constraints=const)
        solution.kinetic_constraints = k_const
        return solution





from functools import partial
from typing import Union, TYPE_CHECKING

from mewpy.io.dto import VariableRecord, DataTransferObject, CompartmentRecord, FunctionTerm
from mewpy.germ.algebra import Expression
from mewpy.germ.models import MetabolicModel
from .engine import Engine
from .engines_utils import build_symbolic, expression_warning, cobra_warning

if TYPE_CHECKING:
    from mewpy.germ.models import RegulatoryModel, Model, MetabolicModel


class CobraModel(Engine):
    def __init__(self, io, config, model=None):
        """
        Engine for COBRApy constraint-based metabolic models
        """
        super().__init__(io, config, model)

    @property
    def model_type(self):
        return 'metabolic'

    @property
    def model(self):

        if self._model is None:
            identifier = self.get_identifier()

            return MetabolicModel(identifier=identifier)

        return self._model

    @staticmethod
    def parse_cobra_objective(objective, model):

        res = {}

        if hasattr(objective, 'expression'):

            for arg in objective.expression.args:

                coef, reaction = str(arg).split('*')

                reaction = model.get(reaction, None)

                if reaction is not None:
                    res[reaction] = coef

        return res

    def get_identifier(self):

        if self.dto.cobra_model:
            return self.dto.cobra_model.id

        return 'model'

    def open(self, mode='r'):

        self._dto = DataTransferObject()

        if not hasattr(self.io, 'reactions'):
            raise OSError(f'{self.io} is not a valid input. Provide a cobrapy model')

        self.dto.cobra_model = self.io

        self.dto.id = self.get_identifier()

        self.dto.name = self.dto.cobra_model.name

    def parse(self):

        if self.dto is None:
            raise OSError('Model is not open')

        if self.dto.id is None:
            raise OSError('Model is not open')

        if self.dto.cobra_model is None:
            raise OSError('Model is not open')

        # -----------------------------------------------------------------------------
        # Compartments
        # -----------------------------------------------------------------------------

        self.dto.compartments = {c_id: CompartmentRecord(id=c_id, name=c_name)
                                 for c_id, c_name in self.dto.cobra_model.compartments.items()}

        # -----------------------------------------------------------------------------
        # Reactions
        # -----------------------------------------------------------------------------
        processed_metabolites = set()
        processed_genes = set()

        for rxn in self.dto.cobra_model.reactions:

            # -----------------------------------------------------------------------------
            # Genes
            # -----------------------------------------------------------------------------

            symbolic, warning = build_symbolic(expression=rxn.gene_reaction_rule)

            if warning:
                self.warnings.append(partial(expression_warning, warning))

            genes = {}

            for symbol in symbolic.atoms(symbols_only=True):
                self.variables[symbol.name].add('gene')

                gene_record = VariableRecord(id=symbol.name,
                                             name=symbol.name,
                                             aliases={symbol.name, symbol.value})

                genes[symbol.name] = gene_record

                processed_genes.add(symbol.name)

                self.dto.genes[symbol.name] = gene_record

            # -----------------------------------------------------------------------------
            # Metabolites
            # -----------------------------------------------------------------------------

            stoichiometry = {}
            metabolites = {}

            for met in rxn.metabolites:
                met_record = VariableRecord(id=met.id,
                                            name=met.name,
                                            aliases={met.id, met.name},
                                            compartment=met.compartment,
                                            charge=met.charge,
                                            formula=met.formula)

                metabolites[met.id] = met_record

                coef = rxn.metabolites[met]

                stoichiometry[met.id] = coef

                processed_metabolites.add(met.id)

                self.variables[met.id].add('metabolite')

                self.dto.metabolites[met.id] = met_record

            # -----------------------------------------------------------------------------
            # GPR Function term
            # -----------------------------------------------------------------------------
            function_term = FunctionTerm(id='gpr_term', symbolic=symbolic, coefficient=1)

            # -----------------------------------------------------------------------------
            # Reaction
            # -----------------------------------------------------------------------------
            reaction_record = VariableRecord(id=rxn.id,
                                             name=rxn.name,
                                             aliases={rxn.id, rxn.name},
                                             bounds=rxn.bounds,
                                             genes=genes,
                                             gpr=function_term,
                                             stoichiometry=stoichiometry,
                                             metabolites=metabolites)

            self.variables[rxn.id].add('reaction')

            self.dto.reactions[rxn.id] = reaction_record

        for met in self.dto.cobra_model.metabolites:

            if met.id not in processed_metabolites:
                met_record = VariableRecord(id=met.id,
                                            name=met.name,
                                            aliases={met.id, met.name},
                                            compartment=met.compartment,
                                            charge=met.charge,
                                            formula=met.formula)

                self.variables[met.id].add('metabolite')

                self.dto.metabolites[met.id] = met_record

        for gene in self.dto.cobra_model.genes:

            if gene.id not in processed_genes:
                gene_record = VariableRecord(id=gene.id,
                                             name=gene.id,
                                             aliases={gene.id})

                self.variables[gene.id].add('gene')

                self.dto.genes[gene.id] = gene_record

    def read(self,
             model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
             variables=None):

        if not model:
            model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = self.model

        if not variables:
            variables = self.variables

        if self.dto.id:
            model._id = self.dto.id

        if self.dto.name:
            model.name = self.dto.name

        model.compartments = {compartment.id: compartment.name
                              for compartment in self.dto.compartments.values()}

        processed_metabolites = set()
        processed_genes = set()

        for rxn_id, rxn_record in self.dto.reactions.items():

            genes = {}

            for gene_id, gene_record in rxn_record.genes.items():

                gene, warning = gene_record.to_variable(model=model,
                                                        types=variables.get(gene_id, {'gene'}),
                                                        name=gene_record.name,
                                                        aliases=gene_record.aliases)

                if warning:
                    self.warnings.append(partial(cobra_warning, warning))

                genes[gene_id] = gene

                processed_genes.add(gene_id)

            stoichiometry = {}

            for met_id, met_record in rxn_record.metabolites.items():

                met, warning = met_record.to_variable(model=model,
                                                      types=variables.get(met_id, {'metabolite'}),
                                                      name=met_record.name,
                                                      aliases=met_record.aliases,
                                                      compartment=met_record.compartment,
                                                      charge=met_record.charge,
                                                      formula=met_record.formula)

                if warning:
                    self.warnings.append(partial(cobra_warning, warning))

                coef = rxn_record.stoichiometry[met_id]

                stoichiometry[met] = coef

                processed_metabolites.add(met_id)

            gpr = Expression(symbolic=rxn_record.gpr.symbolic, variables=genes)

            rxn, warning = rxn_record.to_variable(model=model,
                                                  types=variables.get(rxn_id, {'reaction'}),
                                                  bounds=rxn_record.bounds,
                                                  gpr=gpr,
                                                  stoichiometry=stoichiometry)

            if warning:
                self.warnings.append(partial(cobra_warning, warning))

            model.add(rxn)

        to_append = []

        for met_id, met_record in self.dto.metabolites.items():

            if met_id not in processed_metabolites:
                met, warning = met_record.to_variable(model=model,
                                                      types=variables.get(met_id, {'metabolite'}),
                                                      name=met_record.name,
                                                      aliases=met_record.aliases,
                                                      compartment=met_record.compartment,
                                                      charge=met_record.charge,
                                                      formula=met_record.formula)

                if warning:
                    self.warnings.append(partial(cobra_warning, warning))

                to_append.append(met)

        for gene_id, gene_record in self.dto.genes.items():

            if gene_id not in processed_genes:
                gene, warning = gene_record.to_variable(model=model,
                                                        types=variables.get(gene_id, {'gene'}),
                                                        name=gene_record.name,
                                                        aliases=gene_record.aliases)

                if warning:
                    self.warnings.append(partial(cobra_warning, warning))

                to_append.append(gene)

        model.add(*to_append)

        model.objective = self.parse_cobra_objective(objective=self.dto.cobra_model.objective, model=model)

        return model

    def write(self):
        pass

    def close(self):

        pass

    def clean(self):
        self._dto = None



from typing import TYPE_CHECKING, Union, Tuple

if TYPE_CHECKING:
    from mewpy.germ.variables import Variable


def coefficients_setter(instance: 'Variable', value: Union[Tuple[float, float], Tuple[float], float]):
    """
    Setter for the coefficients attribute of a variable.
    :param instance: the variable instance
    :param value: the value to be set
    :return:
    """
    if value is None:
        value = (0, 1)

    elif isinstance(value, (int, float)):
        value = (value, value)

    elif len(value) == 1:
        value = (value[0], value[0])

    elif len(value) > 1:
        value = tuple(value)

    else:
        raise ValueError('Invalid value for coefficients')

    # if it is a reaction, bounds must be set
    if hasattr(instance, '_bounds'):
        instance._bounds = value

    # if it is a metabolite, the bounds coefficient of the exchange reaction must be returned
    elif hasattr(instance, 'exchange_reaction'):
        if hasattr(instance.exchange_reaction, '_bounds'):
            instance.exchange_reaction._bounds = value

    else:
        instance._coefficients = value

    if instance.model:
        instance.model.notify()



from collections import defaultdict
from typing import TYPE_CHECKING, Any, Union, Generator, Dict, List, Tuple, Set

from .model import Model
from mewpy.util.history import recorder
from mewpy.germ.models.serialization import serialize
from mewpy.util.utilities import generator

if TYPE_CHECKING:
    from mewpy.germ.algebra import Expression
    from mewpy.germ.variables import Gene, Metabolite, Reaction


class MetabolicModel(Model, model_type='metabolic', register=True, constructor=True, checker=True):
    """
    A germ metabolic model consists of a classic Genome-Scale Metabolic (GEM) model,
    containing reactions, metabolites and genes.

    GEM models are systems biology tools used to predict the phenotype of an organism or cellular community
    in range of environmental and genetic conditions.
    To perform phenotype prediction, a metabolic model can be attached to several simulation methods:
        - FBA
        - pFBA
        - FVA
        - ...
    Thus, a germ metabolic model can be associated with a given objective function for the analysis of the model.

    The metabolic model can be loaded with compartments, although these can be inferred from the available
    metabolites.

    A germ metabolic model can hold additional information as follows:
        - demand reactions
        - exchange reactions
        - sink reactions
        - GPRs
        - External compartment

    The metabolic model, as with other models, provides a clean interface for manipulation with the add, remove and
    update methods. One can perform the following operations:
        - Add reactions, metabolites and genes
        - Remove reactions, metabolites and genes
        - Update the objective function
    """
    def __init__(self,
                 identifier: Any,
                 compartments: Dict[str, str] = None,
                 genes: Dict[str, 'Gene'] = None,
                 metabolites: Dict[str, 'Metabolite'] = None,
                 objective: Dict['Reaction', Union[float, int]] = None,
                 reactions: Dict[str, 'Reaction'] = None,
                 **kwargs):

        """
        A germ metabolic model consists of a classic Genome-Scale Metabolic (GEM) model,
        containing reactions, metabolites and genes.

        GEM models are systems biology tools used to predict the phenotype of an organism or cellular community
        in range of environmental and genetic conditions.
        To perform phenotype prediction, a metabolic model can be attached to several simulation methods:
            - FBA
            - pFBA
            - FVA
            - ...
        Thus, a germ metabolic model can be associated with a given objective function for the analysis of the model.

        The metabolic model can be loaded with compartments, although these can be inferred from the available
        metabolites.

        A germ metabolic model can hold additional information as follows:
            - demand reactions
            - exchange reactions
            - sink reactions
            - GPRs
            - External compartment

        The metabolic model, as with other models, provides a clean interface for manipulation with the add, remove and
        update methods. One can perform the following operations:
            - Add reactions, metabolites and genes
            - Remove reactions, metabolites and genes
            - Update the objective function

        :param identifier: identifier, e.g. iMC1010
        :param compartments: a dictionary with additional compartments not encoded in the metabolites
        :param genes: a dictionary with Gene objects. See variables.Gene for more info
        :param metabolites: a dictionary with Metabolite objects. See variables.Metabolite for more info
        :param objective: a dictionary with the Reaction objects that must be considered objective functions of
        the simulations together with the respective coefficients
        :param reactions: a dictionary with Reaction objects. See variables.Reaction for more info
        """
        # compartments attribute can be shared across the children, thus name mangling
        self.__compartments = {}
        self._genes = {}
        self._metabolites = {}
        self._objective = {}
        self._reactions = {}

        super().__init__(identifier,
                         **kwargs)

        # the setters will handle adding and removing variables to the correct containers
        self.compartments = compartments
        self.genes = genes
        self.metabolites = metabolites
        self.objective = objective
        self.reactions = reactions

    # -----------------------------------------------------------------------------
    # Model type manager
    # -----------------------------------------------------------------------------
    @serialize('types', None)
    @property
    def types(self):
        """
        Returns the types of the model
        :return: a set with the types of the model
        """
        _types = {MetabolicModel.model_type}

        _types.update(super(MetabolicModel, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------
    @serialize('genes', 'genes', '_genes')
    @property
    def genes(self) -> Dict[str, 'Gene']:
        """
        It returns a dictionary with the genes of the model. The key is the gene identifier and the value is the
        `Gene` object. To retrieve an iterator with the genes use `yield_genes` method.
        Note that the genes attribute retrieves a copy of the genes' container.
        To update the genes container set new `genes` or use `add` and `remove` methods.
        :return: a dictionary with the genes of the model
        """
        return self._genes.copy()

    @serialize('metabolites', 'metabolites', '_metabolites')
    @property
    def metabolites(self) -> Dict[str, 'Metabolite']:
        """
        It returns a dictionary with the metabolites of the model. The key is the metabolite identifier and the value is
        the `Metabolite` object. To retrieve an iterator with the metabolites use `yield_metabolites` method.
        Note that the metabolites attribute retrieves a copy of the metabolites' container.
        To update the metabolites container set new `metabolites` or use `add` and `remove` methods.
        :return: a dictionary with the metabolites of the model
        """
        return self._metabolites.copy()

    @serialize('objective', 'objective', '_objective')
    @property
    def objective(self) -> Dict['Reaction', Union[float, int]]:
        """
        It returns a dictionary with the objective functions of the model.
        The key is the `Reaction` object and the value is the respective coefficient.
        Note that the objective attribute retrieves a copy of the objective's container.
        To update the objective container set a new `objective` or use `update` method.
        :return: a dictionary with the objective functions of the model
        """
        return self._objective.copy()

    @serialize('reactions', 'reactions', '_reactions')
    @property
    def reactions(self) -> Dict[str, 'Reaction']:
        """
        It returns a dictionary with the reactions of the model. The key is the reaction identifier and the value is
        the `Reaction` object. To retrieve an iterator with the reactions use `yield_reactions` method.
        Note that the reactions attribute retrieves a copy of the reactions' container.
        To update the reactions container set new `reactions` or use `add` and `remove` methods.
        :return: a dictionary with the reactions of the model
        """
        return self._reactions.copy()

    @property
    def compartments(self) -> Dict[str, str]:
        """
        It returns a dictionary with the compartments of the model. The key is the compartment identifier a
        nd the value is the compartment name.
        To retrieve an iterator with the compartments use `yield_compartments` method.
        Note that the compartments attribute retrieves a copy of the compartments' container.
        To update the compartments container set new `compartments`.
        :return:
        """

        compartments = {met.compartment: self.__compartments.get(met.compartment, '')
                        for met in self.yield_metabolites()
                        if met.compartment is not None}

        compartments.update(self.__compartments)

        compartments.update(super(MetabolicModel, self).compartments)

        return compartments

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------
    @compartments.setter
    @recorder
    def compartments(self, value: Dict[str, str]):
        """
        It sets the compartments of the model. The key is the compartment identifier a
        nd the value is the compartment name.
        :param value: a dictionary with the compartments of the model
        :return:
        """
        if not value:
            value = {}

        self.__compartments.update(value)

    @genes.setter
    @recorder
    def genes(self, value: Dict[str, 'Gene']):
        """
        It sets the genes of the model. The key is the gene identifier and the value is the `Gene` object.
        :param value: a dictionary with the genes of the model
        :return:
        """
        if not value:
            value = {}

        self.remove(*self.yield_genes(), history=False)
        self.add(*value.values(), history=False)

    @metabolites.setter
    @recorder
    def metabolites(self, value: Dict[str, 'Metabolite']):
        """
        It sets the metabolites of the model. The key is the metabolite identifier and the value is the `Metabolite`
        object.
        :param value: a dictionary with the metabolites of the model
        :return:
        """
        if not value:
            value = {}

        self.remove(*self.yield_metabolites(), history=False)
        self.add(*value.values(), history=False)

    @objective.setter
    @recorder
    def objective(self, value: Dict['Reaction', Union[float, int]]):
        """
        It sets the objective functions of the model. The key is the `Reaction` object and the value is the respective
        coefficient.
        :param value: a dictionary with the objective functions of the model
        :return:
        """
        if not value:
            value = {}

        if isinstance(value, str):

            value = {self.get(value): 1}

        elif hasattr(value, 'types'):

            value = {value: 1}

        elif isinstance(value, dict):

            value = {self.get(var, var): val for var, val in value.items()}

        else:
            raise ValueError(f'{value} is not a valid objective')

        self._objective = value

        linear_obj = {var.id: coef for var, coef in self._objective.items()}

        for simulator in self.simulators:
            simulator.set_objective(linear=linear_obj, minimize=False)

    @reactions.setter
    @recorder
    def reactions(self, value: Dict[str, 'Reaction']):
        """
        It sets the reactions of the model. The key is the reaction identifier and the value is the `Reaction` object.
        :param value: a dictionary with the reactions of the model
        :return:
        """
        if not value:
            value = {}

        self.remove(*self.yield_reactions(), history=False)
        self.add(*value.values(), history=False)

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def external_compartment(self) -> Union[str, None]:
        """
        It returns the external compartment of the model. This compartment usually corresponds
        to the extracellular space.
        The external compartment is identified as the compartment having more external metabolites.
        External metabolites are often associated with boundary reactions.
        :return: the external compartment of the model
        """
        if not self.compartments:
            return

        if not self._reactions:
            return

        boundary_compartments = defaultdict(int)

        for rxn in self.yield_reactions():

            if rxn.boundary:

                for compartment in rxn.compartments:
                    boundary_compartments[compartment] += 1

        external_compartment = None
        highest_count = 0

        for compartment, count in boundary_compartments.items():

            if count > highest_count:
                external_compartment = compartment
                highest_count = count

        return external_compartment

    def _get_boundaries(self) -> Tuple[Dict[str, 'Reaction'], Dict[str, 'Reaction'], Dict[str, 'Reaction']]:
        """
        It returns the boundary reactions of the model.
        :return: a tuple with exchanges, sinks, demands reactions of the model
        """
        external_compartment = self.external_compartment

        if external_compartment is None:
            return {}, {}, {}

        all_boundaries = [rxn for rxn_id, rxn in self._reactions.items()
                          if rxn.boundary]

        exchanges = {}
        sinks = {}
        demands = {}

        for variable in all_boundaries:

            if variable.types == {'reaction'}:

                if external_compartment in variable.compartments:
                    exchanges[variable.id] = variable

                else:
                    if variable.reversibility:
                        sinks[variable.id] = variable

                    else:
                        demands[variable.id] = variable

        return exchanges, sinks, demands

    @property
    def demands(self) -> Dict[str, 'Reaction']:
        """
        It returns the demand reactions of the model.
        Demand reactions are reactions that consume a metabolite from its compartment.
        :return: a dictionary with the demand reactions of the model
        """
        _, _, demands = self._get_boundaries()
        return demands

    @property
    def exchanges(self) -> Dict[str, 'Reaction']:
        """
        It returns the exchange reactions of the model.
        Exchange reactions are reactions define the environmental conditions of a metabolic model.
        These reactions can provide or consume a metabolite from the extracellular space.
        :return: a dictionary with the exchange reactions of the model
        """
        exchanges, _, _ = self._get_boundaries()
        return exchanges

    @property
    def sinks(self) -> Dict[str, 'Reaction']:
        """
        It returns the sink reactions of the model.
        Sink reactions are reactions that either consume or produce a metabolite in its compartment.
        :return: a dictionary with the sink reactions of the model
        """
        _, sinks, _ = self._get_boundaries()
        return sinks

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------
    def yield_compartments(self) -> Generator[str, None, None]:
        """
        It yields the compartments of the model.
        :return: a generator with the compartments of the model
        """
        return generator(self.compartments)

    def yield_demands(self) -> Generator['Reaction', None, None]:
        """
        It yields the demand reactions of the model.
        :return: a generator with the demand reactions of the model
        """
        return generator(self.demands)

    def yield_exchanges(self) -> Generator['Reaction', None, None]:
        """
        It yields the exchange reactions of the model.
        :return: a generator with the exchange reactions of the model
        """
        return generator(self.exchanges)

    def yield_genes(self) -> Generator['Gene', None, None]:
        """
        It yields the genes of the model.
        :return: a generator with the genes of the model
        """
        return generator(self._genes)

    def yield_gprs(self) -> Generator['Expression', None, None]:
        """
        It yields the GPRs of the model.
        :return: a generator with the GPRs of the model
        """
        return (value.gpr for value in self._reactions.values())

    def yield_metabolites(self) -> Generator['Metabolite', None, None]:
        """
        It yields the metabolites of the model.
        :return: a generator with the metabolites of the model
        """
        return generator(self._metabolites)

    def yield_reactions(self) -> Generator['Reaction', None, None]:
        """
        It yields the reactions of the model.
        :return: a generator with the reactions of the model
        """
        return generator(self._reactions)

    def yield_sinks(self) -> Generator['Reaction', None, None]:
        """
        It yields the sink reactions of the model.
        :return: a generator with the sink reactions of the model
        """
        return generator(self.sinks)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------
    def get(self, identifier: Any, default=None) -> Union['Gene', 'Metabolite', 'Reaction']:
        """
        It returns the object associated with the identifier.
        In case the identifier is not found, it returns the default value.
        For metabolic models, the identifier can be a gene, a metabolite or a reaction.
        :param identifier: the identifier of the object
        :param default: the default value to return in case the identifier is not found
        :return: the object associated with the identifier
        """
        if identifier in self._metabolites:
            return self._metabolites[identifier]

        elif identifier in self._reactions:
            return self._reactions[identifier]

        elif identifier in self._genes:
            return self._genes[identifier]

        else:
            return super(MetabolicModel, self).get(identifier=identifier, default=default)

    def add(self,
            *variables: Union['Gene', 'Metabolite', 'Reaction'],
            comprehensive: bool = True,
            history: bool = True):
        """
        It adds the given variables to the model.
        This method accepts a single variable or a list of variables to be added to specific containers in the model.
        The containers to which the variables will be added are specified by the types.

        For instance, if a variable is simultaneously a metabolite and regulator,
        it will be added to the metabolites and regulators containers.

        If comprehensive is True, the variables and their related variables will be added to the model too.
        If history is True, the changes will be recorded in the history.

        This method notifies all simulators with the recent changes.

        :param variables: the variables to be added to the model
        :param comprehensive: if True, the variables and their related variables will be added to the model too
        :param history: if True, the changes will be recorded in the history
        :return:
        """
        if self.is_a('metabolic'):

            for variable in variables:

                if 'gene' in variable.types:
                    self._add_variable_to_container(variable, '_genes')

                if 'metabolite' in variable.types:
                    self._add_variable_to_container(variable, '_metabolites')

                if 'reaction' in variable.types:
                    if comprehensive:

                        for metabolite in variable.yield_metabolites():
                            self._add_variable_to_container(metabolite, '_metabolites')

                        for gene in variable.yield_genes():
                            self._add_variable_to_container(gene, '_genes')

                    self._add_variable_to_container(variable, '_reactions')

        return super(MetabolicModel, self).add(*variables, comprehensive=comprehensive, history=history)

    def remove(self,
               *variables: Union['Gene', 'Metabolite', 'Reaction'],
               remove_orphans: bool = False,
               history: bool = True):
        """
        It removes the given variables from the model.
        This method accepts a single variable or a list of variables to be removed from specific containers
        in the model.
        The containers from which the variables will be removed are specified by the types.

        For instance, if a variable is simultaneously a metabolite and regulator,
        it will be removed from the metabolites and regulators containers.

        If remove_orphans is True, the variables and their related variables will be removed from the model too.
        If history is True, the changes will be recorded in the history.

        This method notifies all simulators with the recent changes.

        :param variables: the variables to be removed from the model
        :param remove_orphans: if True, the variables and their related variables will be removed from the model too
        :param history: if True, the changes will be recorded in the history
        :return:
        """
        if self.is_a('metabolic'):

            reactions = set()

            for variable in variables:

                if 'gene' in variable.types:
                    self._remove_variable_from_container(variable, '_genes')

                if 'metabolite' in variable.types:
                    self._remove_variable_from_container(variable, '_metabolites')

                if 'reaction' in variable.types:
                    self._remove_variable_from_container(variable, '_reactions')
                    reactions.add(variable)

            if remove_orphans:
                orphan_metabolites = self._get_orphans(to_remove=reactions,
                                                       first_container='metabolites',
                                                       second_container='reactions')

                for metabolite in orphan_metabolites:
                    self._remove_variable_from_container(metabolite, '_metabolites')

                orphan_genes = self._get_orphans(to_remove=reactions,
                                                 first_container='genes',
                                                 second_container='reactions')

                for gene in orphan_genes:
                    self._remove_variable_from_container(gene, '_genes')

        return super(MetabolicModel, self).remove(*variables, remove_orphans=remove_orphans, history=history)

    def update(self,
               compartments: Dict[str, str] = None,
               objective: Dict['Reaction', Union[float, int]] = None,
               variables: Union[List[Union['Gene', 'Metabolite', 'Reaction']],
                                Tuple[Union['Gene', 'Metabolite', 'Reaction']],
                                Set[Union['Gene', 'Metabolite', 'Reaction']]] = None,
               **kwargs):
        """
        It updates the model with relevant information, namely the compartments, objective and variables.

        :param compartments: the compartments to be updated
        :param objective: the objective to be updated
        :param variables: the variables to be updated
        :param kwargs: additional arguments
        :return:
        """
        if compartments is not None:
            self.compartments = compartments

        if variables is not None:
            self.add(*variables)

        if objective is not None:
            self.objective = objective

        super(MetabolicModel, self).update(**kwargs)



from collections import defaultdict
from typing import Union, TYPE_CHECKING, Dict, Tuple, Sequence, Optional

import pandas as pd

from mewpy.util.constants import ModelConstants
from .analysis_utils import run_method_and_decode
from .fba import FBA
from .pfba import pFBA

if TYPE_CHECKING:
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


def slim_fba(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
             objective: Union[str, Dict[str, float]] = None,
             constraints: Dict[str, Tuple[float, float]] = None) -> Optional[float]:
    """
    A Flux Balance Analysis simulation of a metabolic model.
    A slim analysis produces a single and simple solution for the model. This method returns the objective value for the
    FBA simulation.

    Fundamentals of the FBA procedure:
        - A linear problem based on the mass balance constraints
        - Reactions are linear variables constrained by their bounds
        - The objective function is a linear combination of the reactions
        - The objective function is solved using a linear solver

    :param model: a metabolic model to be simulated
    :param objective: the objective function to be used for the simulation.
    If not provided, the default objective is used.
    :param constraints: additional constraints to be used for the simulation.
    :return: the objective value for the simulation
    """
    fba = FBA(model).build()

    objective_value, _ = run_method_and_decode(method=fba, objective=objective, constraints=constraints)
    return objective_value


def slim_pfba(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
              objective: Union[str, Dict[str, float]] = None,
              constraints: Dict[str, Tuple[float, float]] = None) -> Optional[float]:
    """
    A parsimonious Flux Balance Analysis simulation of a metabolic model.
    A slim analysis produces a single and simple solution for the model. This method returns the objective value for the
    pFBA simulation.

    Fundamentals of the pFBA procedure:
        - A linear problem based on the mass balance constraints
        - Reactions are linear variables constrained by their bounds
        - The objective function is a linear combination of the reactions plus the sum of the absolute values of the
        reactions
        - The objective function is solved using a linear solver by minimizing the sum of the absolute values of the
        reactions

    :param model: a metabolic model to be simulated
    If not provided, a new instance will be created.
    :param objective: the objective function to be used for the simulation.
    If not provided, the default objective is used.
    :param constraints: additional constraints to be used for the simulation.
    :return: the objective value for the simulation
    """
    pfba = pFBA(model).build()

    objective_value, _ = run_method_and_decode(method=pfba, objective=objective, constraints=constraints)
    return objective_value


def fva(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
        fraction: float = 1.0,
        reactions: Sequence[str] = None,
        objective: Union[str, Dict[str, float]] = None,
        constraints: Dict[str, Tuple[float, float]] = None) -> pd.DataFrame:
    """
    Flux Variability Analysis (FVA) of a metabolic model.
    FVA is a method to determine the minimum and maximum fluxes for each reaction in a metabolic model.
    It can be used to identify the reactions that are limiting the growth of a cell.
    In MEWpy, FVA is performed by solving a linear problem for each reaction in the model.

    :param model: a metabolic model to be simulated
    :param fraction: the fraction of the optimal solution to be used as the upper bound for the objective function
    (default: 1.0)
    :param reactions: the reactions to be simulated (default: all reactions in the model)
    :param objective: the objective function to be used for the simulation (default: the default objective)
    :param constraints: additional constraints to be used for the simulation (default: None)
    :return: a pandas DataFrame with the minimum and maximum fluxes for each reaction
    """
    if not reactions:
        reactions = model.reactions.keys()

    if not constraints:
        constraints = {}

    if objective:
        if hasattr(objective, 'keys'):
            obj = next(iter(objective.keys()))
        else:
            obj = str(objective)

    else:
        obj = next(iter(model.objective)).id

    _fba = FBA(model).build()
    objective_value, _ = run_method_and_decode(method=_fba, objective=objective, constraints=constraints)
    constraints[obj] = (fraction * objective_value, ModelConstants.REACTION_UPPER_BOUND)

    fba = FBA(model).build()

    result = defaultdict(list)
    for rxn in reactions:
        min_val, _ = run_method_and_decode(method=fba, objective={rxn: 1.0}, constraints=constraints, minimize=True)
        result[rxn].append(min_val)

        max_val, _ = run_method_and_decode(method=fba, objective={rxn: 1.0}, constraints=constraints, minimize=False)
        result[rxn].append(max_val)

    return pd.DataFrame.from_dict(data=result, orient='index', columns=['minimum', 'maximum'])


def single_gene_deletion(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                         genes: Sequence[str] = None,
                         constraints: Dict[str, Tuple[float, float]] = None) -> pd.DataFrame:
    """
    Single gene deletion analysis of a metabolic model.
    Single gene deletion analysis is a method to determine the effect of deleting each gene in a metabolic model.
    It can be used to identify the genes that are essential for the growth of a cell.
    In MEWpy, single gene deletion analysis is performed by solving a linear problem for each gene in the model.
    A gene knockout can switch off reactions associated with the gene, only if the gene is essential for the reaction.

    :param model: a metabolic model to be simulated
    :param genes: the genes to be simulated (default: all genes in the model)
    :param constraints: additional constraints to be used for the simulation (default: None)
    :return: a pandas DataFrame with the fluxes for each gene
    """
    if not constraints:
        constraints = {}

    if not genes:
        genes = model.yield_genes()
    else:
        genes = [model.genes[gene] for gene in genes if gene in model.genes]

    fba = FBA(model).build()
    wt_objective_value, wt_status = run_method_and_decode(method=fba, constraints=constraints)

    state = {gene.id: max(gene.coefficients) for gene in model.yield_genes()}

    result = {}
    for gene in genes:

        gene_coefficient = state.pop(gene.id, 0.0)
        state[gene.id] = 0.0

        gene_constraints = {}
        for reaction in gene.yield_reactions():

            if reaction.gpr.is_none:
                continue

            gpr_eval = reaction.gpr.evaluate(values=state)

            if gpr_eval:
                continue

            gene_constraints[reaction.id] = (0.0, 0.0)

        if gene_constraints:
            solution, status = run_method_and_decode(method=fba, constraints={**constraints, **gene_constraints})
            result[gene.id] = [solution, status]

        else:
            result[gene.id] = [float(wt_objective_value), str(wt_status)]

        state[gene.id] = gene_coefficient

    return pd.DataFrame.from_dict(data=result, orient='index', columns=['growth', 'status'])


def single_reaction_deletion(model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                             reactions: Sequence[str] = None,
                             constraints: Dict[str, Tuple[float, float]] = None) -> pd.DataFrame:
    """
    Single reaction deletion analysis of a metabolic model.
    Single reaction deletion analysis is a method to determine the effect of deleting each reaction
    in a metabolic model.
    It can be used to identify the reactions that are essential for the growth of a cell.
    In MEWpy, single reaction deletion analysis is performed by solving a linear problem for each reaction in the model.

    :param model: a metabolic model to be simulated
    :param reactions: the reactions to be simulated (default: all reactions in the model)
    :param constraints: additional constraints to be used for the simulation (default: None)
    :return: a pandas DataFrame with the fluxes for each reaction
    """
    if not reactions:
        reactions = model.reactions.keys()

    if not constraints:
        constraints = {}

    fba = FBA(model).build()

    result = {}
    for reaction in reactions:
        reaction_constraints = {reaction: (0.0, 0.0)}
        solution, status = run_method_and_decode(method=fba, constraints={**constraints, **reaction_constraints})
        result[reaction] = [solution, status]

    return pd.DataFrame.from_dict(data=result, orient='index', columns=['growth', 'status'])



from math import inf

from mewpy.solvers.solver import VarType
from mewpy.solvers import solver_instance
from mewpy.simulation import get_simulator
from mewpy.simulation.simulation import Simulator
from mewpy.solvers.solution import to_simulation_result
from .. import Preprocessing, ExpressionSet


def iMAT(model, expr, constraints=None, cutoff=(25, 75),
        condition=0, epsilon=1):

    if isinstance(model, Simulator):
        sim = model
    else:
        sim = get_simulator(model)

    if isinstance(expr, ExpressionSet):
        pp = Preprocessing(sim, expr)
        coeffs, _ = pp.percentile(condition, cutoff=cutoff)
        low_coeffs, high_coeffs = coeffs
    else:
        low_coeffs, high_coeffs = expr

    solver = solver_instance(sim)

    if not constraints:
        constraints = {}

    for r_id in sim.reactions:
        lb, _ = sim.get_reaction_bounds(r_id)
        if lb < 0:
            pos, neg = r_id + '+', r_id + '-'
            solver.add_variable(pos, 0, inf, update=False)
            solver.add_variable(neg, 0, inf, update=False)
    solver.update()

    for r_id in sim.reactions:
        lb, _ = sim.get_reaction_bounds(r_id)
        if lb < 0:
            pos, neg = r_id + '+', r_id + '-'
            solver.add_constraint(
                'c' + pos, {r_id: -1, pos: 1}, '>', 0, update=False)
            solver.add_constraint(
                'c' + neg, {r_id: 1, neg: 1}, '>', 0, update=False)
    solver.update()

    objective = list()

    for r_id, val in high_coeffs.items():
        lb, ub = sim.get_reaction_bounds(r_id)
        pos_cons = (lb-epsilon)
        neg_cons = (ub+epsilon)
        pos, neg = 'y_' + r_id + '+', 'y_' + r_id + '-'
        objective.append(pos)
        solver.add_variable(pos, 0, 1, vartype=VarType.BINARY, update=True)
        solver.add_constraint(
            'c' + pos, {r_id: 1, pos: pos_cons}, '>', lb, update=False)
        objective.append(neg)
        solver.add_variable(neg, 0, 1, vartype=VarType.BINARY, update=True)
        solver.add_constraint(
            'c' + neg, {r_id: 1, neg: neg_cons}, '<', ub, update=False)

    solver.update()

    for r_id, val in low_coeffs.items():
        lb, ub = sim.get_reaction_bounds(r_id)
        x_var = 'x_' + r_id
        objective.append(x_var)
        solver.add_variable(x_var, 0, 1, vartype=VarType.BINARY, update=True)
        solver.add_constraint(
            'c' + x_var + '_pos', {r_id: 1, x_var: lb}, '>', lb, update=False)
        solver.add_constraint(
            'c' + x_var + '_neg', {r_id: 1, x_var: ub}, '<', ub, update=False)

    solver.update()

    object = {x: 1 for x in objective}

    solution = solver.solve(object, minimize=False, constraints=constraints)
    
    res = to_simulation_result(model, None, constraints, sim, solution)
    return res


from mewpy.simulation import get_simulator
from mewpy.simulation.simulation import Simulator
from .. import Preprocessing, ExpressionSet


def eFlux(model, expr, condition=0, scale_rxn=None, scale_value=1,
          constraints=None, parsimonious=False, max_exp = None, **kwargs):
    """ Run an E-Flux simulation (Colijn et al, 2009).

    Arguments:
        model: a REFRAMED or COBRApy model or a MEWpy Simulator.
        expr (ExpressionSet): transcriptomics data.
        condition: the condition to use in the simulation\
                (default:0, the first condition is used if more than one.)
            scale_rxn (str): reaction to scale flux vector (optional)
        scale_value (float): scaling factor (mandatory if scale_rxn\
            is specified)
        constraints (dict): additional constraints (optional)
        parsimonious (bool): compute a parsimonious solution (default: False)

    Returns:
        Solution: solution
    """
    if isinstance(model, Simulator):
        sim = model
    else:
        sim = get_simulator(model)

    if isinstance(expr, ExpressionSet):
        pp = Preprocessing(sim, expr)
        rxn_exp = pp.reactions_expression(condition)
    else:
        rxn_exp = expr

    if max_exp is None:
        max_exp = max(rxn_exp.values())

    bounds = {}

    for r_id in sim.reactions:
        val = rxn_exp[r_id] / max_exp if r_id in rxn_exp else 1
        lb, ub = sim.get_reaction_bounds(r_id)
        lb2 = -val if lb < 0 else 0
        ub2 = val if ub > 0 else 0
        bounds[r_id] = (lb2, ub2)

    if constraints:
        for r_id, x in constraints.items():
            lb, ub = x if isinstance(x, tuple) else (x, x)
            lb2 = -1 if lb < 0 else 0
            ub2 = 1 if ub > 0 else 0
            bounds[r_id] = (lb2, ub2)

    if parsimonious:
        sol = sim.simulate(constraints=bounds, method='pFBA')
    else:
        sol = sim.simulate(constraints=bounds)

    if scale_rxn is not None:

        if sol.fluxes[scale_rxn] != 0:
            k = abs(scale_value / sol.fluxes[scale_rxn])
        else:
            k = 0

        for r_id, val in sol.fluxes.items():
            sol.fluxes[r_id] = val * k

    return sol



from typing import Any, TYPE_CHECKING, Dict, Union, Generator, Tuple

import pandas as pd

from mewpy.germ.algebra import Expression, parse_expression
from mewpy.util.utilities import generator
from mewpy.germ.models.serialization import serialize
from mewpy.util.history import recorder
from mewpy.io.engines.engines_utils import expression_warning
from .variable import Variable, variables_from_symbolic


# Preventing circular dependencies that only happen due to type checking
if TYPE_CHECKING:
    from .regulator import Regulator
    from .target import Target
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


class Interaction(Variable, variable_type='interaction', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 target: 'Target' = None,
                 regulatory_events: Dict[Union[float, int], Expression] = None,
                 **kwargs):

        """
        A regulatory interaction is regularly associated with a target and the regulatory
        events modelling the coefficients of this target variable.

        The regulatory events determine the multiple coefficients that the target variable can take
        depending on the logic and conditions encoded into an expression object.

        The expression must hold the multiple variables/regulators that participate in the regulatory event/expression.

        The set of regulators is inferred from the regulatory events.
        Also, the regulatory events are used to generate a regulatory table,
        namely the possible coefficients of the target for the default (or not) coefficients of the regulators.

        :param identifier: the interaction identifier
        :param target: the target variable that is regulated by the regulators/expressions logic
        :param regulatory_events: A dictionary comprehending coefficient-expression pairs.
        The expression when evaluated to true results into the corresponding coefficient
        that must be associated with the target coefficients
        """

        if not target:
            target = None

        if not regulatory_events:
            regulatory_events = {0.0: Expression()}

        self._regulatory_events = {}
        self._target = None

        super().__init__(identifier,
                         **kwargs)

        # it handles addition of a target. It fulfills the correct containers/attributes
        self.add_target(target, history=False)

        # it handles setting a regulatory event. It fulfills the correct containers/attributes for the regulators
        # associated with the expression og a given event
        for coefficient, expression in regulatory_events.items():
            self.add_regulatory_event(coefficient=coefficient, expression=expression, history=False)

        # There is an alternative constructor for building an interaction with a stringify-like expression rule.
        # Parsing takes the hard job to infer the correct relation between regulators

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------
    @property
    def types(self):
        # noinspection PyUnresolvedReferences
        _types = {Interaction.variable_type}

        _types.update(super(Interaction, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------
    def __str__(self):

        if self.target:
            target_str = f'{self.target.id} || '

        else:
            target_str = ''

        expression_str = ' + '.join([f'{coefficient} = {expression.to_string()}'
                                     for coefficient, expression in self.regulatory_events.items()
                                     if not expression.is_none])

        return target_str + expression_str

    def _interaction_to_html(self):
        """
        It returns a html dict representation.
        """
        html_dict = {'Target': self.target,
                     'Regulators': ', '.join(self.regulators),
                     'Regulatory events': ', '.join([f'{coefficient} = {expression.to_string()}'
                                                     for coefficient, expression in self.regulatory_events.items()
                                                     if not expression.is_none])}
        return html_dict

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------
    @serialize('regulatory_events', 'regulatory_events', '_regulatory_events')
    @property
    def regulatory_events(self) -> Dict[Union[float, int], Expression]:
        """
        A dictionary comprehending coefficient-expression key-value pairs.
        The expression when evaluated to true results into the corresponding coefficient that must be associated
        with the target coefficients
        """
        return self._regulatory_events.copy()

    @serialize('target', 'target', '_target')
    @property
    def target(self) -> 'Target':
        """
        The target variable that is regulated by the regulators/expressions logic
        """
        return self._target

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------
    @regulatory_events.setter
    @recorder
    def regulatory_events(self, value: Dict[Union[float, int], Expression]):
        """
        Setting a dictionary comprehending coefficient-expression key-value pairs
        to replace the current regulatory events.
        This setter adds and removes regulatory events.
        It also handles the linear problems associated to this interaction
        """
        if not value:
            value = {0.0: Expression()}

        for coefficient, expression in self.regulatory_events.items():
            self.remove_regulatory_event(coefficient=coefficient,
                                         remove_orphans=True, history=False)

        for coefficient, expression in value.items():
            self.add_regulatory_event(coefficient=coefficient, expression=expression, history=False)

    @target.setter
    @recorder
    def target(self, value: 'Target'):
        """
        Setting a new Target variable.
        This setter adds and removes target variable.
        It also handles the linear problems associated to this interaction
        """
        self.remove_target(remove_from_model=True, history=False)
        self.add_target(value, history=False)

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------
    @property
    def regulators(self) -> Dict[str, 'Regulator']:
        """
        It returns a dictionary with the regulators associated to this interaction
        """
        return {reg_id: regulator
                for expression in self.yield_expressions()
                for reg_id, regulator in expression.variables.items()}

    @property
    def regulatory_truth_table(self) -> pd.DataFrame:
        """
        It calculates the truth table for this interaction based on the regulatory events.
        The truth table is composed by the combination of values taken by empty, numeric and symbolic variables
        available in the regulatory events of this interaction.
        That is, the regulatory table is a pandas DataFrame with the possible coefficients
        of the target variable for the default (or not) coefficients of the regulators.

        :return: It returns a pandas DataFrame with the truth table
        """
        truth_tables = []

        for coefficient, expression in self._regulatory_events.items():
            df = expression.truth_table(strategy='max', coefficient=coefficient)

            df.index = [self.target.id if self.target else 'result'] * df.shape[0]

            truth_tables.append(df)

        regulatory_events = pd.concat(truth_tables)

        return regulatory_events

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------
    def yield_regulatory_events(self) -> Generator[Tuple[Union[float, int], Expression], None, None]:
        """
        It yields all regulatory events
        """
        return ((coefficient, expression) for coefficient, expression in self._regulatory_events.items())

    def yield_expressions(self) -> Generator[Expression, None, None]:
        """
        It yields all expressions
        """
        return generator(self._regulatory_events)

    def yield_regulators(self) -> Generator['Regulator', None, None]:
        """
        It yields all regulators
        """
        return generator(self.regulators)

    # -----------------------------------------------------------------------------
    # Polymorphic constructors
    # -----------------------------------------------------------------------------
    @classmethod
    def from_string(cls,
                    identifier: Any,
                    rule: str,
                    target: 'Target',
                    coefficient: Union[float, int] = 1.0,
                    model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                    **kwargs) -> 'Interaction':
        """
        A regulatory interaction is regularly associated with a target and the regulatory
        events modelling the coefficients of this target variable.

        A regulatory interaction can be assembled from a string object encoding an algebra expression

        :param identifier: The interaction identifier
        :param rule: A string object encoding an algebra expression
        :param target: The target variable that is regulated by the regulators/expressions logic
        :param coefficient: the coefficient that should be taken by the target
        whether this expression is evaluated positively
        :param model: If a model is provided, this interaction will be associated to the model
        :param kwargs: Additional arguments
        :return: A new interaction object
        """
        try:

            symbolic = parse_expression(rule)

            regulators = variables_from_symbolic(symbolic=symbolic, types=('regulator',), model=model)

            expression = Expression(symbolic=symbolic, variables=regulators)

        except SyntaxError as exc:

            expression_warning(f'{rule} cannot be parsed')

            raise exc

        instance = cls(identifier=identifier,
                       target=target,
                       regulatory_events={coefficient: expression},
                       model=model,
                       **kwargs)

        return instance

    @classmethod
    def from_expression(cls,
                        identifier: Any,
                        expression: Expression,
                        target: 'Target',
                        coefficient: Union[float, int] = 1.0,
                        model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
                        **kwargs) -> 'Interaction':

        """
        A regulatory interaction is regularly associated with a target and the regulatory
        events modelling the coefficients of this target variable.

        A regulatory interaction can be assembled from an expression object encoding an algebra expression

        :param identifier: The interaction identifier
        :param expression: An expression object encoding an algebra expression
        :param target: The target variable that is regulated by the regulators/expressions logic
        :param coefficient: the coefficient that should be taken by the target
        whether this expression is evaluated positively
        :param model: If a model is provided, this interaction will be associated to the model
        :param kwargs: Additional arguments
        :return: A new interaction object
        """

        if not isinstance(expression, Expression):
            raise TypeError(f'expression must be an {Expression} object')

        instance = cls(identifier=identifier,
                       target=target,
                       regulatory_events={coefficient: expression},
                       model=model,
                       **kwargs)

        return instance

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------
    def update(self,
               regulatory_events: Dict[Union[float, int], Expression] = None,
               target: 'Target' = None,
               **kwargs):
        """
        It performs an update operation to this interaction.
        The update operation is similar to a dictionary update.

        It also handles the linear problems associated to this interaction

        Note that, some update operations are not registered in history.
        It is strongly advisable to use update outside history context manager

        :param target: the target variable that is regulated by the regulators/expressions logic
        :param regulatory_events: A dictionary comprehending coefficient-expression pairs.
        The expression when evaluated to true results into the corresponding coefficient
        that must be associated with the target coefficients
        :param kwargs: Other arguments for the base variable, such as identifier, name, etc
        :return:
        """
        super(Interaction, self).update(**kwargs)

        if target is not None:
            self.target: 'Target' = target

        if regulatory_events is not None:
            self.regulatory_events: Dict[Union[float, int], Expression] = regulatory_events

    def add_target(self, target: 'Target', history=True):
        """
        It adds a new target to this regulatory interaction.
        If a target is already associated with this interaction, the current target will be removed and replaced by the
        new target.

        It also handles the linear problems associated to this interaction

        This operation can be reverted using the model history
        :param target: the target variable that is regulated by the regulators/expressions logic
        :param history: Whether to register this operation in the model history
        """
        if history:
            self.history.queue_command(undo_func=self.remove_target,
                                       undo_kwargs={'remove_from_model': True,
                                                    'history': False},
                                       func=self.add_target,
                                       kwargs={'target': target,
                                               'history': history})

        if self.target and not target:
            return self.remove_target(remove_from_model=True, history=False)

        elif not self.target and not target:
            return

        else:
            self._target = target
            self.target._interaction = self

            if self.model:
                self.model.add(self.target, comprehensive=False, history=False)

                self.model.notify()

    def remove_target(self, remove_from_model=True, history=True):
        """
        It removes the current target of this regulatory interaction.

        It also handles the linear problems associated to this interaction

        This operation can be reverted using the model history
        :param remove_from_model: Whether this operation should be reflected in the model too.
        It is usually advisable to remove a target from the interaction and model at the same time
        :param history: Whether to register this operation in the model history
        """
        if history:
            self.history.queue_command(undo_func=self.add_target,
                                       undo_kwargs={'target': self.target,
                                                    'history': False},
                                       func=self.remove_target,
                                       kwargs={'remove_from_model': remove_from_model,
                                               'history': history})

        if self.target:
            target = self.target

            self.target._interaction = None
            self._target = None

            if self.model:

                if remove_from_model:
                    self.model.remove(target, remove_orphans=False, history=False)

                self.model.notify()

    def add_regulatory_event(self,
                             coefficient: Union[float, int],
                             expression: Expression,
                             history=True):
        """
        It adds a new regulatory event to this interaction.
        If a regulatory event with the same coefficient is already associated with this interaction, the current
        regulatory event will be removed and replaced by the new regulatory event.

        It also handles the linear problems associated to this interaction

        This operation can be reverted using the model history
        :param coefficient: the coefficient that should be taken by the target for this expression
        :param expression: An expression object encoding an algebra expression
        :param history: Whether to register this operation in the model history
        :return: None
        """

        if not isinstance(expression, Expression):
            raise TypeError(f'expression must be an {Expression} object. '
                            f'To set None, provide an empty Expression()')

        if history:
            self.history.queue_command(undo_func=self.remove_regulatory_event,
                                       undo_kwargs={'coefficient': coefficient,
                                                    'expression': expression,
                                                    'remove_orphans': True,
                                                    'history': False},
                                       func=self.add_regulatory_event,
                                       kwargs={'coefficient': coefficient,
                                               'expression': expression,
                                               'history': history})

        to_add = []

        for regulator in expression.variables.values():
            regulator.update(interactions={self.id: self},
                             model=self.model)

            if self.model:
                if regulator.id not in self.model.regulators:
                    to_add.append(regulator)

        self._regulatory_events[coefficient] = expression

        if self.model:
            self.model.add(*to_add, comprehensive=False, history=False)

            self.model.notify()

    def remove_regulatory_event(self,
                                coefficient: Union[float, int],
                                remove_orphans=True,
                                history=True):
        """
        It removes a regulatory event from this interaction.
        If the expression is not associated with the coefficient, nothing will happen.

        It also handles the linear problems associated to this interaction

        This operation can be reverted using the model history
        :param coefficient: the coefficient that should be taken by the target for this expression
        :param remove_orphans: Whether to remove orphaned regulators from the model
        :param history: Whether to register this operation in the model history
        :return:
        """
        expression = self.regulatory_events.get(coefficient)
        if expression is None:
            raise ValueError(f'No regulatory event associated with coefficient {coefficient}')

        if not isinstance(expression, Expression):
            raise TypeError(f'expression must be an {Expression} object. '
                            f'To set None, provide an empty Expression()')

        if history:
            self.history.queue_command(undo_func=self.add_regulatory_event,
                                       undo_kwargs={'coefficient': coefficient,
                                                    'expression': expression,
                                                    'history': False},
                                       func=self.remove_regulatory_event,
                                       kwargs={'coefficient': coefficient,
                                               'remove_orphans': remove_orphans,
                                               'history': history})

        to_remove = []

        for regulator in expression.variables.values():

            # noinspection PyProtectedMember
            del regulator._interactions[self.id]

            # noinspection PyProtectedMember
            if not regulator._interactions:
                to_remove.append(regulator)

        self._regulatory_events.pop(coefficient)

        if self.model:

            if remove_orphans:
                self.model.remove(*to_remove, remove_orphans=False, history=False)

            self.model.notify()



from typing import Tuple

import numpy as np
import pandas as pd
from itertools import combinations

from pandas._typing import FilePathOrBuffer

from ..simulation import get_simulator
from ..simulation.simulation import Simulator
from ..util.parsing import Boolean, GeneEvaluator, build_tree


class ExpressionSet:

    def __init__(self, identifiers: list, conditions: list,
                 expression: np.array, p_values: np.array = None):
        """Expression set. The expression values are a numpy array with shape
        (len(identifiers) x len(conditions)).

        Args:
            identifiers (list): Gene or Proteins identifiers
            conditions (list): Time, experiment,... identifiers.
            expression (np.array): expression values.
            p_values (np.array, optional): p-values. Defaults to None.
        """
        n = len(identifiers)
        m = len(conditions)
        if expression.shape != (n, m):
            raise ValueError(
                f"The shape of the expression {expression.shape} does not "
                f"match the expression and conditions sizes ({n},{m})")

        self._identifiers = identifiers
        self._identifier_index = {iden: idx for idx, iden in
                                  enumerate(identifiers)}
        self._conditions = [str(x) for x in conditions]
        self._condition_index = {cond: idx for idx, cond in
                                 enumerate(self._conditions)}
        self._expression = expression
        self._p_values = p_values

    def shape(self):
        """Returns:
            (tuple): the Expression dataset shape
        """
        return self._expression.shape

    def __getitem__(self, item):
        """
        Index the ExpressionSet.
        """
        return self._expression.__getitem__(item)

    def get_condition(self, condition=None, **kwargs):
        """

        Args:
            condition ([type], optional): [description]. Defaults to None.

        Returns:
            [type]: [description]
        """

        if isinstance(condition, int):
            values = self[:, condition]
        elif isinstance(condition, str):
            values = self[:, self._condition_index[condition]]
        else:
            values = self[:, :]

        # format
        form = kwargs.get('format', 'dict')
        if form and condition is not None:
            if form == 'list':
                return values.tolist()
            elif form == 'dict':
                return dict(zip(self._identifiers, values.tolist()))
            else:
                return values
        else:
            return values

    @classmethod
    def from_dataframe(cls, data_frame):
        """Read expression data from a pandas.DataFrame.

        Args:
            data_frame (Dataframe): The expression Dataframe

        Returns:
            ExpressionSet: the expression dataset from the dataframe.
        """
        # pandas columns might be integers only wich will cause problems in the p-value verification step
        columns = [str(x) for x in data_frame.columns]
        data_frame.columns = columns

        conditions = [c for c in columns if "p-value" not in c]
        p_value_keys = [c for c in columns if "p-value" in c]
        if p_value_keys:
            p_values = data_frame[p_value_keys].values
        else:
            p_values = None

        expression = data_frame[conditions].values
        identifiers = data_frame.index.tolist()
        return ExpressionSet(identifiers, conditions, expression, p_values)

    @classmethod
    def from_csv(cls, file_path: FilePathOrBuffer, **kwargs):
        """Read expression data from a comma separated values (csv) file.

        Args:
            file_path (str): the csv file path.

        Returns:
            ExpressionSet: the expression dataset from the csv file.
        """
        data = pd.read_csv(file_path, **kwargs)
        return cls.from_dataframe(data)

    @property
    def dataframe(self):
        """Build a pandas.DataFrame from the ExpressionProfile.
        Columns headers are conditions and
        line indexes identifiers (genes/proteins)
        """

        if self._p_values is None:
            expression = self._expression
            conditions = self._conditions
        else:
            expression = np.concatenate((self._expression, self.p_values),
                                        axis=1)
            conditions = self._conditions + self.p_value_columns

        return pd.DataFrame(expression,
                            index=self._identifiers,
                            columns=conditions)

    @property
    def p_value_columns(self):
        """ Generate the p-value column names."""
        return [f"{c[0]} {c[1]} p-value"
                for c in combinations(self._conditions, 2)]

    @property
    def p_values(self):
        """Returns the numpy array of p-values.

        Raises:
            ValueError: [description]
        """
        if not self._p_values.all():
            raise ValueError("No p-values defined.")
        else:
            return self._p_values

    @p_values.setter
    def p_values(self, p_values: np.array):
        """Sets p-values

        Args:
            p_values (np.array): [description]

        Raises:
            ValueError: [description]
        """
        if p_values is not None:
            if p_values.shape[1] != len(self.p_value_columns):
                raise ValueError("p-values do not cover all conditions")

        self._p_values = p_values

    @p_values.deleter
    def p_values(self):
        """Delete p_values."""
        self._p_values = None

    def differences(self, p_value=0.005):
        """Calculate the differences based on the MADE method.

        Args:
            p_value (float, optional): [description]. Defaults to 0.005.

        Returns:
            dict: A dictionary of differences
        """

        diff = {}
        for idx, iden in enumerate(self._identifiers):
            diff[iden] = []
            for i in range(1, len(self._conditions)):
                start, end = self._expression[idx, i - 1: i + 1]
                p_val = self.p_values[idx, i - 1]
                if p_val <= p_value:
                    if start < end:
                        diff[iden].append(+1)
                    elif start > end:
                        diff[iden].append(-1)
                    else:
                        diff[iden].append(0)
                else:
                    diff[iden].append(0)
        return diff

    def minmax(self, condition=None):
        """ Return the min and max values for the specified condition.

        Args:
            condition (str): str or int or None, optional (default None)
            The condition to obtain the min and max values for.

        Returns
        -------
        tuple of (min, max)

        """
        values = self.get_condition(condition)
        return np.amin(values), np.amax(values)

    def apply(self, function: None):
        """Apply a function to all expression values.

        :param function: the unary function to be applyied. Default log base 2.
        :type function: callable
        """
        if function is None:
            import math
            def function(x): return math.log(x, 2)
        f = np.vectorize(function)
        self._expression = f(self._expression)

    def quantile_pipeline(self,
                          missing_values: float = None,
                          n_neighbors: int = 5,
                          weights: str = 'uniform',
                          metric: str = 'nan_euclidean',
                          n_quantiles: int = None,
                          q: float = 0.33) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Quantile preprocessing pipeline. It performs the following steps:
            1. KNN imputation of missing values
            2. Quantile transformation
            3. Quantile binarization
        :param missing_values: The placeholder for the missing values.
        All occurrences of missing_values will be imputed.
        :param n_neighbors: Number of neighboring samples to use for imputation.
        :param weights: Weight function used in prediction. Possible values:
            - 'uniform': uniform weights. All points in each neighborhood are weighted equally.
            - 'distance': weight points by the inverse of their distance.
            in this case, closer neighbors of a query point
        :param metric: Metric used to compute the distance between samples. The default metric is nan_euclidean,
        which is the euclidean distance ignoring missing values. Consult sklearn documentation for more information.
        :param n_quantiles: Number of quantiles to be computed.
        It corresponds to the number of landmarks used to discretize
        :param q: Quantile to compute
        :return: Quantile preprocessed expression matrix, quantile expression binarized matrix
        """
        expression = knn_imputation(self._expression,
                                    missing_values=missing_values,
                                    n_neighbors=n_neighbors,
                                    weights=weights,
                                    metric=metric)
        expression = quantile_transformation(expression, n_quantiles=n_quantiles)

        binary_expression = quantile_binarization(expression, q=q)

        expression = pd.DataFrame(expression, self._identifiers, self._conditions)
        binary_expression = pd.DataFrame(binary_expression, self._identifiers, self._conditions)
        return expression, binary_expression


# ----------------------------------------------------------------------------------------------------------------------
# Preprocessing using KNNImputer and Quantile transformation/binarization
# ----------------------------------------------------------------------------------------------------------------------
def knn_imputation(expression: np.ndarray,
                   missing_values: float = None,
                   n_neighbors: int = 5,
                   weights: str = 'uniform',
                   metric: str = 'nan_euclidean') -> np.ndarray:
    """
    KNN imputation of missing values in the expression matrix. It uses the scikit-learn KNNImputer (Consult sklearn
    documentation for more information).
    The default metric is nan_euclidean, which is the euclidean distance ignoring missing values.

    :param expression: Expression matrix
    :param missing_values: The placeholder for the missing values. All occurrences of missing_values will be imputed.
    :param n_neighbors: Number of neighboring samples to use for imputation.
    :param weights: Weight function used in prediction. Possible values:
        - 'uniform': uniform weights. All points in each neighborhood are weighted equally.
        - 'distance': weight points by the inverse of their distance. in this case, closer neighbors of a query point
    :param metric: Metric used to compute the distance between samples. The default metric is nan_euclidean, which is
    the euclidean distance ignoring missing values. Consult sklearn documentation for more information.
    :return: Imputed expression matrix
    """
    try:
        # noinspection PyPackageRequirements
        from sklearn.impute import KNNImputer
    except ImportError:
        raise ImportError('The package scikit-learn is not installed. '
                          'To preprocess gene expression data, please install scikit-learn (pip install scikit-learn).')

    if missing_values is None:
        missing_values = np.nan

    if n_neighbors is None:
        n_neighbors = 5

    if weights is None:
        weights = 'uniform'

    if metric is None:
        metric = 'nan_euclidean'

    imputation = KNNImputer(missing_values=missing_values,
                            n_neighbors=n_neighbors,
                            weights=weights,
                            metric=metric)

    return imputation.fit_transform(expression)


def quantile_transformation(expression: np.ndarray, n_quantiles: int = None) -> np.ndarray:
    """
    Quantile transformation of the expression matrix. It uses the scikit-learn quantile_transform (Consult sklearn
    documentation for more information).
    :param expression: Expression matrix
    :param n_quantiles: Number of quantiles to be computed. It corresponds to the number of landmarks used to discretize
    :return: Quantile transformed expression matrix
    """
    try:
        # noinspection PyPackageRequirements
        from sklearn.preprocessing import quantile_transform
    except ImportError:
        raise ImportError('The package scikit-learn is not installed. '
                          'To preprocess gene expression data, please install scikit-learn (pip install scikit-learn).')

    if n_quantiles is None:
        n_quantiles = expression.shape[1]

    return quantile_transform(expression, n_quantiles=n_quantiles, axis=0, random_state=0)


def quantile_binarization(expression: np.ndarray, q: float = 0.33) -> np.ndarray:
    """
    It computes the q-th quantile of the expression matrix using np.quantile (consult numpy documentation for more
    information). Then, it binarizes the expression matrix using the threshold computed.
    :param expression: Expression matrix
    :param q: Quantile to compute
    :return: Binarized expression matrix
    """
    threshold = np.quantile(expression, q)

    threshold_mask = expression >= threshold
    expression[threshold_mask] = 1
    expression[~threshold_mask] = 0
    return expression


def gene_to_reaction_expression(model, gene_exp, and_func=min, or_func=max):
    """Process reaction level from GPRs

    Args:
        model: A model or a MEWpy Simulation
        gene_exp (dict): gene identifiers and expression values
        and_func ([type], optional): Function for AND. Defaults to min.
        or_func ([type], optional): Function for OR. Defaults to max.

    Returns:
        dict: Reaction levels
    """
    if isinstance(model, Simulator):
        sim = model
    else:
        sim = get_simulator(model)

    rxn_exp = {}
    evaluator = GeneEvaluator(gene_exp, and_func, or_func, unexpressed_value=None)
    for rxn_id in sim.reactions:
        gpr = sim.get_gpr(rxn_id)
        if gpr:
            tree = build_tree(gpr, Boolean)
            op_set = tree.get_operands().intersection(set(gene_exp.keys()))
            if len(op_set) == 0:
                lv = None
            else:
                lv = tree.evaluate(evaluator.f_operand, evaluator.f_operator)
            rxn_exp[rxn_id] = lv
    return rxn_exp


class Preprocessing:
    """Formulation and implementation of preprocessing decisions.
        (A) Types of gene mapping methods
        (B) Types of thresholding approaches (global and local).
        (C) Formulation of combinations of number of states (Global, Local)
        (D) Decisions about the order in which thresholding and gene mapping
        are performed.
        For Order 1, gene expression is converted to reaction activity followed
        by thresholding of reaction activity;
        For Order 2, thresholding ofgene expression is followed by its
        conversion to reaction activity.

        [1]Anne Richelle,Chintan Joshi,Nathan E. Lewis, Assessing key decisions
        for transcriptomic data integration in biochemical networks, PLOS, 2019
        https://doi.org/10.1371/journal.pcbi.1007185
    """

    def __init__(self, model: Simulator, data: ExpressionSet, **kwargs):
        """[summary]

        Args:
            model (Simulator): [description]
            data (ExpressionSet): [description]
            and_func (function): (optional)
            or_func (function): (optional)
        """
        self.model = model
        self.data = data
        self._conf = kwargs

    def reactions_expression(self, condition, and_func=None, or_func=None):
        exp = self.data.get_condition(condition, format='dict')
        and_func = self._conf.get(
            'and_func', min) if and_func is None else and_func
        or_func = self._conf.get(
            'or_func', max) if or_func is None else or_func
        rxn_exp = gene_to_reaction_expression(
            self.model, exp, and_func, or_func)
        # Removes None if maybe is none to evaluate GPRs
        res = {k: v for k, v in rxn_exp.items() if v is not None}
        return res

    def percentile(self, condition=None, cutoff=25):
        """Processes a percentil threshold and returns the respective
        reaction coefficients, ie, a dictionary of reaction:coeff

        Args:
            condition ([type], optional): [description]. Defaults to None.
            cutoff (int, optional): [description]. Defaults to 25.

        Returns:
            dict, float: the coefficients and threshold
        """
        if type(cutoff) is tuple:
            coef = []
            thre = []
            for cut in cutoff:
                rxn_exp = self.reactions_expression(condition)
                threshold = np.percentile(list(rxn_exp.values()), cut)
                coeffs = {r_id: threshold - val for r_id,
                                                    val in rxn_exp.items() if val < threshold}
                coef.append(coeffs)
                thre.append(threshold)
            coeffs = tuple(coef)
            threshold = tuple(thre)
        else:
            rxn_exp = self.reactions_expression(condition)
            threshold = np.percentile(list(rxn_exp.values()), cutoff)
            coeffs = {r_id: threshold - val for r_id,
                                                val in rxn_exp.items() if val < threshold}
        return coeffs, threshold



"""scipy ODE solver"""
from .ode import ODEMethod, ODESolver
from scipy.integrate._ivp import solve_ivp

methods = {
    ODEMethod.RK45: 'RK45',
    ODEMethod.RK23: 'RK23',
    ODEMethod.DOP853: 'DOP853',
    ODEMethod.Radau: 'Radau',
    ODEMethod.BDF: 'BDF',
    ODEMethod.LSODA: 'LSODA',
}


class ScipySolver(ODESolver):

    def __init__(self, func, method='LSODA'):
        self.func = func
        self.method = method
        self.initial_condition = None

    def set_initial_condition(self, initial_condition):
        self.initial_condition = initial_condition

    def solve(self, y0, t_points, **kwargs):
        t_span=[t_points[0],t_points[-1]]
        sol = solve_ivp(self.func, t_span, y0, method=methods[self.method], t_eval=t_points,**kwargs)
        C = [c[-1] for c in sol.y]
        return C, sol.t, sol.y



from enum import Enum

from .boolean_csv import BooleanRegulatoryCSV
from . co_expression_csv import CoExpressionRegulatoryCSV
from .target_regulator_csv import TargetRegulatorRegulatoryCSV
from .cobra_model import CobraModel
from .reframed_model import ReframedModel
from .json import JSON
from .metabolic_sbml import MetabolicSBML
from .regulatory_sbml import RegulatorySBML


class Engines(Enum):
    """
    List of all engines to read/write files/models

    """

    BooleanRegulatoryCSV = BooleanRegulatoryCSV
    CoExpressionRegulatoryCSV = CoExpressionRegulatoryCSV
    TargetRegulatorRegulatoryCSV = TargetRegulatorRegulatoryCSV
    RegulatorySBML = RegulatorySBML
    MetabolicSBML = MetabolicSBML
    CobraModel = CobraModel
    ReframedModel = ReframedModel
    JSON = JSON

    @classmethod
    def has_engine(cls, engine):
        try:
            return cls[engine.__name__]

        except KeyError:

            return False

    @classmethod
    def get(cls, engine, default=None):
        try:
            return cls[engine.__name__]

        except KeyError:

            return default



from mewpy.util.utilities import Singleton


class MEWPYSolvers(Singleton):

    def __init__(self):
        self._mewpy_solvers = dict()

    def build(self):
        try:
            from .gurobi_solver import GurobiSolver
            self._mewpy_solvers['gurobi'] = GurobiSolver
        except ImportError:
            pass

        try:
            from .cplex_solver import CplexSolver
            self._mewpy_solvers['cplex'] = CplexSolver
        except ImportError:
            pass

        try:
            from .optlang_solver import OptLangSolver
            self._mewpy_solvers['optlang'] = OptLangSolver
        except ImportError:
            pass

    def get_solvers(self):
        if not self._mewpy_solvers:
            self.build()
        return self._mewpy_solvers


class MEWPYODESolvers(Singleton):

    def __init__(self):
        self._mewpy_ode_solvers = dict()

    def build(self):
        try:
            from .scikits_solver import ScikitsODESolver
            self._mewpy_ode_solvers['scikits'] = ScikitsODESolver
        except ImportError:
            pass

        try:
            from .scipy_solver import ScipySolver
            self._mewpy_ode_solvers['scipy'] = ScipySolver
        except ImportError:
            pass

        try:
            from .odespy_solver import ODESpySolver
            self._mewpy_ode_solvers['odespy'] = ODESpySolver
        except ImportError:
            pass

    def get_solvers(self):
        if not self._mewpy_ode_solvers:
            self.build()
        return self._mewpy_ode_solvers


__MEWPY_solvers__ = MEWPYSolvers().get_solvers()
__MEWPY_ode_solvers__ = MEWPYODESolvers().get_solvers()







from typing import Union, Dict, Tuple, List
from warnings import warn

from mewpy.germ.analysis import FBA
from mewpy.solvers import Solution
from mewpy.solvers.solver import Solver
from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel
from mewpy.util.constants import ModelConstants
from mewpy.germ.solution import ModelSolution, DynamicSolution


def _get_boolean_state_from_reaction_flux(flux_rate: float) -> bool:
    # A reaction can have different flux values between two solutions,
    # though the state remains the same. Thus, non-zero flux stands for ON state
    # while zero flux stands for OFF state
    if abs(flux_rate) > ModelConstants.TOLERANCE:
        return True

    return False


def _find_duplicated_state(state, regulatory_solution, regulatory_reactions, regulatory_metabolites):
    mask = []
    for regulator, value in regulatory_solution.items():

        state_value = state[regulator]

        if regulator in regulatory_reactions or regulator in regulatory_metabolites:
            state_value = _get_boolean_state_from_reaction_flux(state_value)
            value = _get_boolean_state_from_reaction_flux(value)

        mask.append(value == state_value)

    return all(mask)


class RFBA(FBA):

    def __init__(self,
                 model: Union[Model, MetabolicModel, RegulatoryModel],
                 solver: Union[str, Solver, None] = None,
                 build: bool = False,
                 attach: bool = False):
        """
        Regulatory Flux Balance Analysis (RFBA) of a metabolic-regulatory model.
        Implementation of a steady-state and dynamic versions of RFBA for a integrated metabolic-regulatory model.

        For more details consult Covert et al. 2004 at https://doi.org/10.1038/nature02456

        :param model: a MetabolicModel, RegulatoryModel or GERM model. The model is used to retrieve
        variables and constraints to the linear problem
        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used, but it will be overwritten
        if build is true.
        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """
        self._regulatory_reactions = []
        self._regulatory_metabolites = []
        super().__init__(model=model, solver=solver, build=build, attach=attach)

    def _build(self):
        """
        It builds the linear problem for RFBA. It is a linear problem with the following constraints:
            - Metabolic constraints

        The regulatory layer is not considered in the linear problem. It is only considered in the simulation step.
        :return:
        """
        if self.model.is_metabolic():
            self._build_mass_constraints()

            self._regulatory_reactions = [rxn.id
                                          for rxn in self.model.yield_reactions()
                                          if rxn.is_regulator()]

            self._regulatory_metabolites = [met.id
                                            for met in self.model.yield_metabolites()
                                            if met.is_regulator()]

            self._linear_objective = {var.id: value for var, value in self.model.objective.items()}
            self._minimize = False

    def initial_state(self, state: Dict[str, float] = None) -> Dict[str, float]:
        """
        Method responsible for retrieving the initial state of the model.
        The initial state is the state of all regulators found in the Metabolic-Regulatory model.
        :param state: the initial state of the model
        :return: dict of regulatory/metabolic variable keys (regulators) and a value of 0 or 1
        """
        if not state:
            state = {}

        if not self.model.is_regulatory():
            return state

        initial_state = {}
        for regulator in self.model.yield_regulators():
            if regulator.id in state:
                initial_state[regulator.id] = state[regulator.id]

            elif regulator.is_metabolite() and regulator.exchange_reaction:
                if regulator.exchange_reaction.id in state:
                    initial_state[regulator.id] = state[regulator.exchange_reaction.id]

                else:
                    initial_state[regulator.id] = abs(regulator.exchange_reaction.lower_bound)

            else:
                initial_state[regulator.id] = max(regulator.coefficients)

        return initial_state

    def decode_regulatory_state(self, state: Dict[str, float]) -> Dict[str, float]:
        """
        It solves the boolean regulatory network for a given specific state.
        It also updates all targets having a valid regulatory interaction associated with it for the resulting state

        :param state: dict of regulatory variable keys (regulators) and a value of 0, 1 or float
        (reactions and metabolites predicates)
        :return: dict of target keys and a value of the resulting state
        """
        if not self.model.is_regulatory():
            return {}

        # Solving regulatory model synchronously for the regulators according to the initial state
        # Targets are associated with a single regulatory interaction
        result = {}
        for interaction in self.model.yield_interactions():

            # solving regulators state only
            if not interaction.target.is_regulator():
                continue

            # an interaction can have multiple regulatory events, namely one for 0 and another for 1
            for coefficient, event in interaction.regulatory_events.items():
                if event.is_none:
                    continue

                eval_result = event.evaluate(values=state)
                if eval_result:
                    result[interaction.target.id] = coefficient
                else:
                    result[interaction.target.id] = 0.0
        return result

    def decode_metabolic_state(self, state: Dict[str, float]) -> Dict[str, float]:
        """
        It solves the boolean regulatory network for a given specific state.
        It also updates all targets having a valid regulatory interaction associated with it for the resulting state

        :param state: dict of regulatory variable keys (regulators) and a value of 0, 1 or float
        (reactions and metabolites predicates)
        :return: dict of target keys and a value of the resulting state
        """
        if not self.model.is_regulatory():
            return {}

        # Solving the whole regulatory model synchronously, as asynchronously would take too much time
        # Targets are associated with a single regulatory interaction
        result = {}
        for interaction in self.model.yield_interactions():

            # an interaction can have multiple regulatory events, namely one for 0 and another for 1
            for coefficient, event in interaction.regulatory_events.items():
                if event.is_none:
                    continue

                eval_result = event.evaluate(values=state)
                if eval_result:
                    result[interaction.target.id] = coefficient
                else:
                    result[interaction.target.id] = 0.0
        return result

    def decode_constraints(self, state: Dict[str, float]) -> Dict[str, Tuple[float, float]]:
        """
        Method responsible for decoding the RFBA metabolic state, namely the state of all metabolic genes associated
        at least with one reaction in the GPRs rule.

        :param state: dict of regulatory/metabolic variable keys (metabolic target) and a value of 0 or 1
        :return: dict of constraints of the resulting metabolic state
        """
        # This method retrieves the constraints associated with a given metabolic/regulatory state

        if not self.model.is_metabolic():
            return {}

        constraints = {}
        for rxn in self.model.yield_reactions():

            if rxn.gpr.is_none:
                continue

            res = rxn.gpr.evaluate(values=state)

            if not res:
                constraints[rxn.id] = (0.0, 0.0)

        return constraints

    def next_state(self, state: Dict[str, float], solver_kwargs: Dict = None) -> Tuple[Dict[str, float],
                                                                                       Solution]:
        """
        Retrieves the next state for the provided state

        Solves the boolean regulatory model using method regulatory_simulation(state) and decodes the metabolic state
        for that state or initial state (according to the flag regulatory_state)

        :param state: dict of regulatory/metabolic variable keys (regulatory and metabolic target) and a value of 0, 1
        or float (reactions and metabolites predicates)
        :param solver_kwargs: solver kwargs
        :return: dict of all regulatory/metabolic variables keys and a value of the resulting state
        """
        state = state.copy()

        if not solver_kwargs:
            solver_kwargs = {}

        if 'constraints' in solver_kwargs:
            constraints = solver_kwargs['constraints'].copy()
        else:
            constraints = {}

        # Regulatory state from a synchronous boolean simulation
        # noinspection PyTypeChecker
        regulatory_state = self.decode_regulatory_state(state=state)

        # Next state is the previous state plus the regulatory state
        next_state = {**state, **regulatory_state}

        # After a simulation of the regulators outputs, the state of the targets are retrieved now
        metabolic_state = self.decode_metabolic_state(state=next_state)

        regulatory_constraints = self.decode_constraints(metabolic_state)
        regulatory_constraints = {**constraints, **regulatory_constraints}

        solver_kwargs['constraints'] = regulatory_constraints
        solver_solution = self.solver.solve(**solver_kwargs)

        if solver_solution.values:
            solver_solution.values = {**metabolic_state, **solver_solution.values}
        else:
            solver_solution.values = {**metabolic_state, **{rxn: 0.0 for rxn in self.model.reactions}}

        # update the next state with the regulatory/metabolic state
        for reaction in self._regulatory_reactions:
            next_state[reaction] = solver_solution.values.get(reaction, 0.0)

        for metabolite in self._regulatory_metabolites:

            reaction = self.model.metabolites[metabolite].exchange_reaction

            if reaction:
                next_state[metabolite] = solver_solution.values.get(reaction.id, 0.0)

        return next_state, solver_solution

    def _dynamic_optimize(self,
                          initial_state: Dict[str, float] = None,
                          iterations: int = 10,
                          to_solver: bool = False,
                          solver_kwargs: Dict = None):
        """
        RFBA model dynamic simulation (until the metabolic-regulatory state is reached).

        First, the boolean regulatory model is solved using the initial state. If there is no initial state ,
        the state of all regulatory variables is inferred from the model.

        At the same time, the metabolic state is also decoded from the initial state of all metabolic genes.
        If otherwise set by using the regulatory flag, the metabolic state is decoded from the boolean regulatory model
        simulation.

        The result of the boolean regulatory model updates the state of all targets, while the result of the
        metabolic state decoding updates the state of all reactions and metabolites that are also regulatory variables.

        Then, this new resulting metabolic-regulatory state is used to solve again the boolean regulatory model and
        decode again the metabolic state towards the retrieval of a new regulatory and metabolic states.

        This cycle is iterated until a given state is repeated, the so called metabolic-regulatory steady-state.
        Hence, dynamic RFBA is based on finding the cyclic attractors of the metabolic-regulatory networks
        Alternatively, an iteration limit may be reached.

        Finally, all states between the cyclic attractor are used for decoding final metabolic states using
        the resulting metabolic-regulatory state. Thus, a solution/simulation is obtained for each mid-state in the
        cyclic attractor of the metabolic-regulatory networks

        Objective and constraint-based model analysis method (pFBA, FBA, ...) can be altered here reversibly.

        :param initial_state: a dictionary of variable ids and their values to set as initial state
        :param iterations: The maximum number of iterations. Default: 10
        :param to_solver: Whether to return the solution as a SolverSolution instance. Default: False
        :param solver_kwargs: Keyword arguments to pass to the solver.
        :return: A DynamicSolution instance or a list of solver Solutions if to_solver is True.
        """
        if not initial_state:
            initial_state = {}

        if not solver_kwargs:
            solver_kwargs = {}

        # It takes the initial state from the model and then updates with the initial state provided as input
        initial_state = self.initial_state(initial_state)

        regulatory_solutions = []
        solver_solutions = []

        # solve using the initial state
        # noinspection PyTypeChecker
        state, solver_solution = self.next_state(state=initial_state, solver_kwargs=solver_kwargs)
        regulatory_solutions.append(state)
        solver_solutions.append(solver_solution)

        i = 1
        steady_state = False
        while not steady_state:
            # Updating state upon state. See next state for further detail
            state, solver_solution = self.next_state(state=state, solver_kwargs=solver_kwargs)

            for regulatory_solution in regulatory_solutions:

                is_duplicated = _find_duplicated_state(state=state, regulatory_solution=regulatory_solution,
                                                       regulatory_reactions=self._regulatory_reactions,
                                                       regulatory_metabolites=self._regulatory_metabolites)
                if not is_duplicated:
                    continue

                steady_state = True
                break

            # add the new state to the list of regulatory solutions
            regulatory_solutions.append(state)
            solver_solutions.append(solver_solution)

            # if the maximum number of iterations is reached, the simulation is stopped
            if i < iterations:
                i += 1

            else:
                def iteration_limit(message):
                    warn(message, UserWarning, stacklevel=2)

                iteration_limit("Iteration limit reached")
                steady_state = True

        if to_solver:
            return solver_solutions

        minimize = solver_kwargs.get('minimize', self._minimize)
        solutions = [ModelSolution.from_solver(method=self.method,
                                               solution=solver_solution,
                                               model=self.model,
                                               minimize=minimize)
                     for solver_solution in solver_solutions]
        return DynamicSolution(*solutions)

    def _steady_state_optimize(self,
                               initial_state: Dict[str, float] = None,
                               to_solver: bool = False,
                               solver_kwargs: Dict = None) -> Union[ModelSolution, Solution]:
        """
        RFBA model one-step simulation (pseudo steady-state).

        First, the boolean regulatory model is simulated using the initial state. If there is no initial state ,
        the state of all regulatory variables is inferred from the model.

        At the same time, the metabolic state is also decoded from the initial state of all metabolic genes.
        If otherwise set by using the regulatory flag, the metabolic state is decoded from the boolean regulatory model
        simulation.

        The result of the boolean regulatory model updates the state of all targets, while the result of the
        metabolic state decoding updates the state of all reactions and metabolites that are also regulatory variables.

        Then, this new state is used to decode the final metabolic state using the resulting metabolic-regulatory state.

        Objective and constraint-based model analysis method (pFBA, FBA, ...) can be altered here reversibly.

        :param initial_state: a dictionary of variable ids and their values to set as initial state
        :param to_solver: Whether to return the solution as a SolverSolution instance. Default: False
        :param solver_kwargs: A dictionary with the solver arguments. Default: None
        :return: A ModelSolution instance or a SolverSolution instance if to_solver is True.
        """
        if not initial_state:
            initial_state = {}

        if not solver_kwargs:
            solver_kwargs = {}

        if 'constraints' in solver_kwargs:
            constraints = solver_kwargs['constraints'].copy()
        else:
            constraints = {}

        # It takes the initial state from the model and then updates with the initial state provided as input
        initial_state = self.initial_state(initial_state)

        # Regulatory state from a synchronous boolean simulation
        # noinspection PyTypeChecker
        regulatory_state = self.decode_regulatory_state(state=initial_state)

        # After a simulation of the regulators outputs, the state of the targets are retrieved now
        metabolic_state = self.decode_metabolic_state(state={**initial_state, **regulatory_state})

        regulatory_constraints = self.decode_constraints(metabolic_state)
        metabolic_regulatory_constraints = {**constraints, **regulatory_constraints}

        solver_kwargs['constraints'] = metabolic_regulatory_constraints
        solution = self.solver.solve(**solver_kwargs)

        if solution.values:
            solution.values = {**initial_state, **regulatory_state, **solution.values}

        if to_solver:
            return solution

        minimize = solver_kwargs.get('minimize', self._minimize)
        return ModelSolution.from_solver(method=self.method, solution=solution, model=self.model, minimize=minimize)

    def _optimize(self,
                  initial_state: Dict[str, float] = None,
                  dynamic: bool = False,
                  iterations: int = 10,
                  to_solver: bool = False,
                  solver_kwargs: Dict = None,
                  **kwargs) -> Union[DynamicSolution, ModelSolution, List[Solution]]:
        """
        RFBA simulation.

        First, the boolean regulatory model is solved using the initial state. If there is no initial state ,
        the state of all regulatory variables is inferred from the model.

        At the same time, the metabolic state is also decoded from the initial state of all metabolic genes.
        If otherwise set by using the regulatory flag, the metabolic state is decoded from the boolean regulatory model
        simulation.

        The result of the boolean regulatory model updates the state of all targets, while the result of the
        metabolic state decoding updates the state of all reactions and metabolites that are also regulatory variables.

        Then, this new resulting metabolic-regulatory state is used to solve again the boolean regulatory model and
        decode again the metabolic state towards the retrieval of a new regulatory and metabolic states.

        If the dynamic flag is set to True, the simulation will continue until the metabolic state remains the same
        between two consecutive simulations, the so called metabolic-regulatory steady-state.
        Hence, dynamic RFBA is based on finding the cyclic attractors of the metabolic-regulatory networks
        Alternatively, an iteration limit may be reached.

        Finally, all states between the cyclic attractor are used for decoding final metabolic states using
        the resulting metabolic-regulatory state. Thus, a solution/simulation is obtained for each mid-state in the
        cyclic attractor of the metabolic-regulatory networks

        Objective and constraint-based model analysis method (pFBA, FBA, ...) can be altered here reversibly.

        :param initial_state: a dictionary of variable ids and their values to set as initial state
        :param dynamic: If True, the model is simulated until the metabolic-regulatory steady-state is reached.
        :param iterations: The maximum number of iterations. Default: 10
        :param to_solver: Whether to return the solution as a SolverSolution instance. Default: False
        :param solver_kwargs: Keyword arguments to pass to the solver. See LinearProblem.optimize for details.
        :return: A DynamicSolution instance or a list of solver Solutions if to_solver is True.
        """
        if dynamic:
            return self._dynamic_optimize(initial_state=initial_state, iterations=iterations,
                                          to_solver=to_solver, solver_kwargs=solver_kwargs)

        return self._steady_state_optimize(initial_state=initial_state,
                                           to_solver=to_solver, solver_kwargs=solver_kwargs)

    def optimize(self,
                 to_solver: bool = False,
                 solver_kwargs: Dict = None,
                 initial_state: Dict[str, float] = None,
                 dynamic: bool = False,
                 iterations: int = 10,
                 **kwargs) -> Union[DynamicSolution, ModelSolution, Solution, List[Solution]]:
        """
        RFBA simulation.

        First, the boolean regulatory model is solved using the initial state. If there is no initial state ,
        the state of all regulatory variables is inferred from the model.

        At the same time, the metabolic state is also decoded from the initial state of all metabolic genes.
        If otherwise set by using the regulatory flag, the metabolic state is decoded from the boolean regulatory model
        simulation.

        The result of the boolean regulatory model updates the state of all targets, while the result of the
        metabolic state decoding updates the state of all reactions and metabolites that are also regulatory variables.

        Then, this new resulting metabolic-regulatory state is used to solve again the boolean regulatory model and
        decode again the metabolic state towards the retrieval of a new regulatory and metabolic states.

        If the dynamic flag is set to True, the simulation will continue until the metabolic state remains the same
        between two consecutive simulations, the so called metabolic-regulatory steady-state.
        Hence, dynamic RFBA is based on finding the cyclic attractors of the metabolic-regulatory networks
        Alternatively, an iteration limit may be reached.

        Finally, all states between the cyclic attractor are used for decoding final metabolic states using
        the resulting metabolic-regulatory state. Thus, a solution/simulation is obtained for each mid-state in the
        cyclic attractor of the metabolic-regulatory networks

        Objective and constraint-based model analysis method (pFBA, FBA, ...) can be altered here reversibly.

        :param to_solver: Whether to return the solution as a SolverSolution instance. Default: False
        :param solver_kwargs: Keyword arguments to pass to the solver. See LinearProblem.optimize for details.
        :param initial_state: a dictionary of variable ids and their values to set as initial state
        :param dynamic: If True, the model is simulated until the metabolic-regulatory steady-state is reached.
        :param iterations: The maximum number of iterations. Default: 10
        :return: A DynamicSolution instance or a list of solver Solutions if to_solver is True.
        """
        # build solver if out of sync
        if not self.synchronized:
            self.build()

        if not solver_kwargs:
            solver_kwargs = {}

        # concrete optimize
        return self._optimize(to_solver=to_solver, solver_kwargs=solver_kwargs,
                              initial_state=initial_state, dynamic=dynamic, iterations=iterations,
                              **kwargs)
